# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===========================================================================
"""Writer for storing imported summary event data to a SQLite DB."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os
import sys
import time

import six

from tensorboard.compat import tf
from tensorboard.util import tb_logging
from tensorboard.util import tensor_util


logger = tb_logging.get_logger()

# Struct bundling a tag with its SummaryMetadata and a list of values, each of
# which are a tuple of step, wall time (as a float), and a TensorProto.
TagData = collections.namedtuple("TagData", ["tag", "metadata", "values"])


class SqliteWriter(object):
    """Sends summary data to SQLite using python's sqlite3 module."""

    def __init__(self, db_connection_provider):
        """Constructs a SqliteWriterEventSink.

        Args:
          db_connection_provider: Provider function for creating a DB connection.
        """
        self._db = db_connection_provider()

    def _make_blob(self, bytestring):
        """Helper to ensure SQLite treats the given data as a BLOB."""
        # Special-case python 2 pysqlite which uses buffers for BLOB.
        if sys.version_info[0] == 2:
            return buffer(bytestring)  # noqa: F821 (undefined name)
        return bytestring

    def _create_id(self):
        """Returns a freshly created DB-wide unique ID."""
        cursor = self._db.cursor()
        cursor.execute("INSERT INTO Ids DEFAULT VALUES")
        return cursor.lastrowid

    def _maybe_init_user(self):
        """Returns the ID for the current user, creating the row if needed."""
        user_name = os.environ.get("USER", "") or os.environ.get("USERNAME", "")
        cursor = self._db.cursor()
        cursor.execute(
            "SELECT user_id FROM Users WHERE user_name = ?", (user_name,)
        )
        row = cursor.fetchone()
        if row:
            return row[0]
        user_id = self._create_id()
        cursor.execute(
            """
            INSERT INTO USERS (user_id, user_name, inserted_time)
            VALUES (?, ?, ?)
            """,
            (user_id, user_name, time.time()),
        )
        return user_id

    def _maybe_init_experiment(self, experiment_name):
        """Returns the ID for the given experiment, creating the row if needed.

        Args:
          experiment_name: name of experiment.
        """
        user_id = self._maybe_init_user()
        cursor = self._db.cursor()
        cursor.execute(
            """
            SELECT experiment_id FROM Experiments
            WHERE user_id = ? AND experiment_name = ?
            """,
            (user_id, experiment_name),
        )
        row = cursor.fetchone()
        if row:
            return row[0]
        experiment_id = self._create_id()
        # TODO: track computed time from run start times
        computed_time = 0
        cursor.execute(
            """
            INSERT INTO Experiments (
              user_id, experiment_id, experiment_name,
              inserted_time, started_time, is_watching
            ) VALUES (?, ?, ?, ?, ?, ?)
            """,
            (
                user_id,
                experiment_id,
                experiment_name,
                time.time(),
                computed_time,
                False,
            ),
        )
        return experiment_id

    def _maybe_init_run(self, experiment_name, run_name):
        """Returns the ID for the given run, creating the row if needed.

        Args:
          experiment_name: name of experiment containing this run.
          run_name: name of run.
        """
        experiment_id = self._maybe_init_experiment(experiment_name)
        cursor = self._db.cursor()
        cursor.execute(
            """
            SELECT run_id FROM Runs
            WHERE experiment_id = ? AND run_name = ?
            """,
            (experiment_id, run_name),
        )
        row = cursor.fetchone()
        if row:
            return row[0]
        run_id = self._create_id()
        # TODO: track actual run start times
        started_time = 0
        cursor.execute(
            """
            INSERT INTO Runs (
              experiment_id, run_id, run_name, inserted_time, started_time
            ) VALUES (?, ?, ?, ?, ?)
            """,
            (experiment_id, run_id, run_name, time.time(), started_time),
        )
        return run_id

    def _maybe_init_tags(self, run_id, tag_to_metadata):
        """Returns a tag-to-ID map for the given tags, creating rows if needed.

        Args:
          run_id: the ID of the run to which these tags belong.
          tag_to_metadata: map of tag name to SummaryMetadata for the tag.
        """
        cursor = self._db.cursor()
        # TODO: for huge numbers of tags (e.g. 1000+), this is slower than just
        # querying for the known tag names explicitly; find a better tradeoff.
        cursor.execute(
            "SELECT tag_name, tag_id FROM Tags WHERE run_id = ?", (run_id,)
        )
        tag_to_id = {
            row[0]: row[1]
            for row in cursor.fetchall()
            if row[0] in tag_to_metadata
        }
        new_tag_data = []
        for tag, metadata in six.iteritems(tag_to_metadata):
            if tag not in tag_to_id:
                tag_id = self._create_id()
                tag_to_id[tag] = tag_id
                new_tag_data.append(
                    (
                        run_id,
                        tag_id,
                        tag,
                        time.time(),
                        metadata.display_name,
                        metadata.plugin_data.plugin_name,
                        self._make_blob(metadata.plugin_data.content),
                    )
                )
        cursor.executemany(
            """
            INSERT INTO Tags (
              run_id, tag_id, tag_name, inserted_time, display_name, plugin_name,
              plugin_data
            ) VALUES (?, ?, ?, ?, ?, ?, ?)
            """,
            new_tag_data,
        )
        return tag_to_id

    def write_summaries(self, tagged_data, experiment_name, run_name):
        """Transactionally writes the given tagged summary data to the DB.

        Args:
          tagged_data: map from tag to TagData instances.
          experiment_name: name of experiment.
          run_name: name of run.
        """
        logger.debug("Writing summaries for %s tags", len(tagged_data))
        # Connection used as context manager for auto commit/rollback on exit.
        # We still need an explicit BEGIN, because it doesn't do one on enter,
        # it waits until the first DML command - which is totally broken.
        # See: https://stackoverflow.com/a/44448465/1179226
        with self._db:
            self._db.execute("BEGIN TRANSACTION")
            run_id = self._maybe_init_run(experiment_name, run_name)
            tag_to_metadata = {
                tag: tagdata.metadata
                for tag, tagdata in six.iteritems(tagged_data)
            }
            tag_to_id = self._maybe_init_tags(run_id, tag_to_metadata)
            tensor_values = []
            for tag, tagdata in six.iteritems(tagged_data):
                tag_id = tag_to_id[tag]
                for step, wall_time, tensor_proto in tagdata.values:
                    dtype = tensor_proto.dtype
                    shape = ",".join(
                        str(d.size) for d in tensor_proto.tensor_shape.dim
                    )
                    # Use tensor_proto.tensor_content if it's set, to skip relatively
                    # expensive extraction into intermediate ndarray.
                    data = self._make_blob(
                        tensor_proto.tensor_content
                        or tensor_util.make_ndarray(tensor_proto).tobytes()
                    )
                    tensor_values.append(
                        (tag_id, step, wall_time, dtype, shape, data)
                    )
            self._db.executemany(
                """
                INSERT OR REPLACE INTO Tensors (
                  series, step, computed_time, dtype, shape, data
                ) VALUES (?, ?, ?, ?, ?, ?)
                """,
                tensor_values,
            )


# See tensorflow/contrib/tensorboard/db/schema.cc for documentation.
_SCHEMA_STATEMENTS = [
    """
    CREATE TABLE IF NOT EXISTS Ids (
      id INTEGER PRIMARY KEY
    )
    """,
    """
    CREATE TABLE IF NOT EXISTS Descriptions (
      id INTEGER PRIMARY KEY,
      description TEXT
    )
    """,
    """
    CREATE TABLE IF NOT EXISTS Tensors (
      rowid INTEGER PRIMARY KEY,
      series INTEGER,
      step INTEGER,
      dtype INTEGER,
      computed_time REAL,
      shape TEXT,
      data BLOB
    )
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS
      TensorSeriesStepIndex
    ON
      Tensors (series, step)
    WHERE
      series IS NOT NULL
      AND step IS NOT NULL
    """,
    """
    CREATE TABLE IF NOT EXISTS TensorStrings (
      rowid INTEGER PRIMARY KEY,
      tensor_rowid INTEGER NOT NULL,
      idx INTEGER NOT NULL,
      data BLOB
    )
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS TensorStringIndex
    ON TensorStrings (tensor_rowid, idx)
    """,
    """
    CREATE TABLE IF NOT EXISTS Tags (
      rowid INTEGER PRIMARY KEY,
      run_id INTEGER,
      tag_id INTEGER NOT NULL,
      inserted_time DOUBLE,
      tag_name TEXT,
      display_name TEXT,
      plugin_name TEXT,
      plugin_data BLOB
    )
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS TagIdIndex
    ON Tags (tag_id)
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS
      TagRunNameIndex
    ON
      Tags (run_id, tag_name)
    WHERE
      run_id IS NOT NULL
      AND tag_name IS NOT NULL
    """,
    """
    CREATE TABLE IF NOT EXISTS Runs (
      rowid INTEGER PRIMARY KEY,
      experiment_id INTEGER,
      run_id INTEGER NOT NULL,
      inserted_time REAL,
      started_time REAL,
      finished_time REAL,
      run_name TEXT
    )
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS RunIdIndex
    ON Runs (run_id)
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS RunNameIndex
    ON Runs (experiment_id, run_name)
    WHERE run_name IS NOT NULL
    """,
    """
    CREATE TABLE IF NOT EXISTS Experiments (
      rowid INTEGER PRIMARY KEY,
      user_id INTEGER,
      experiment_id INTEGER NOT NULL,
      inserted_time REAL,
      started_time REAL,
      is_watching INTEGER,
      experiment_name TEXT
    )
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS ExperimentIdIndex
    ON Experiments (experiment_id)
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS ExperimentNameIndex
    ON Experiments (user_id, experiment_name)
    WHERE experiment_name IS NOT NULL
    """,
    """
    CREATE TABLE IF NOT EXISTS Users (
      rowid INTEGER PRIMARY KEY,
      user_id INTEGER NOT NULL,
      inserted_time REAL,
      user_name TEXT,
      email TEXT
    )
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS UserIdIndex
    ON Users (user_id)
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS UserNameIndex
    ON Users (user_name)
    WHERE user_name IS NOT NULL
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS UserEmailIndex
    ON Users (email)
    WHERE email IS NOT NULL
    """,
    """
    CREATE TABLE IF NOT EXISTS Graphs (
      rowid INTEGER PRIMARY KEY,
      run_id INTEGER,
      graph_id INTEGER NOT NULL,
      inserted_time REAL,
      graph_def BLOB
    )
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS GraphIdIndex
    ON Graphs (graph_id)
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS GraphRunIndex
    ON Graphs (run_id)
    WHERE run_id IS NOT NULL
    """,
    """
    CREATE TABLE IF NOT EXISTS Nodes (
      rowid INTEGER PRIMARY KEY,
      graph_id INTEGER NOT NULL,
      node_id INTEGER NOT NULL,
      node_name TEXT,
      op TEXT,
      device TEXT,
      node_def BLOB
    )
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS NodeIdIndex
    ON Nodes (graph_id, node_id)
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS NodeNameIndex
    ON Nodes (graph_id, node_name)
    WHERE node_name IS NOT NULL
    """,
    """
    CREATE TABLE IF NOT EXISTS NodeInputs (
      rowid INTEGER PRIMARY KEY,
      graph_id INTEGER NOT NULL,
      node_id INTEGER NOT NULL,
      idx INTEGER NOT NULL,
      input_node_id INTEGER NOT NULL,
      input_node_idx INTEGER,
      is_control INTEGER
    )
    """,
    """
    CREATE UNIQUE INDEX IF NOT EXISTS NodeInputsIndex
    ON NodeInputs (graph_id, node_id, idx)
    """,
]


# Defines application ID as hexspeak for "TBOARD[0]", with an expansion digit.
# This differs from 0xFEEDABEE used in schema.h, which unfortunately exceeds
# the range of a signed 32-bit int and thus gets interpreted as 0.
_TENSORBOARD_APPLICATION_ID = 0x7B0A12D0


# Arbitrary user-controlled version number.
_TENSORBOARD_USER_VERSION = 0


def initialize_schema(connection):
    """Initializes the TensorBoard sqlite schema using the given connection.

    Args:
      connection: A sqlite DB connection.
    """
    cursor = connection.cursor()
    cursor.execute(
        "PRAGMA application_id={}".format(_TENSORBOARD_APPLICATION_ID)
    )
    cursor.execute("PRAGMA user_version={}".format(_TENSORBOARD_USER_VERSION))
    with connection:
        for statement in _SCHEMA_STATEMENTS:
            lines = statement.strip("\n").split("\n")
            message = lines[0] + ("..." if len(lines) > 1 else "")
            logger.debug("Running DB init statement: %s", message)
            cursor.execute(statement)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Classes and functions for handling the ListSessionGroups API call."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import operator
import re

import six

from google.protobuf import struct_pb2

from tensorboard.data import provider
from tensorboard.plugins.hparams import api_pb2
from tensorboard.plugins.hparams import error
from tensorboard.plugins.hparams import metadata
from tensorboard.plugins.hparams import metrics


class Handler(object):
    """Handles a ListSessionGroups request."""

    def __init__(self, context, experiment_id, request):
        """Constructor.

        Args:
          context: A backend_context.Context instance.
          experiment_id: A string, as from `plugin_util.experiment_id`.
          request: A ListSessionGroupsRequest protobuf.
        """
        self._context = context
        self._experiment_id = experiment_id
        self._request = request
        self._extractors = _create_extractors(request.col_params)
        self._filters = _create_filters(request.col_params, self._extractors)
        # Query for all Hparams summary metadata up front to minimize calls to
        # the underlying DataProvider.
        self._hparams_run_to_tag_to_content = context.hparams_metadata(
            experiment_id
        )
        # Since an context.experiment() call may search through all the runs, we
        # cache it here.
        self._experiment = context.experiment_from_metadata(
            experiment_id, self._hparams_run_to_tag_to_content
        )

    def run(self):
        """Handles the request specified on construction.

        Returns:
          A ListSessionGroupsResponse object.
        """
        session_groups = self._build_session_groups()
        session_groups = self._filter(session_groups)
        self._sort(session_groups)
        return self._create_response(session_groups)

    def _build_session_groups(self):
        """Returns a list of SessionGroups protobuffers from the summary
        data."""

        # Algorithm: We keep a dict 'groups_by_name' mapping a SessionGroup name
        # (str) to a SessionGroup protobuffer. We traverse the runs associated with
        # the plugin--each representing a single session. We form a Session
        # protobuffer from each run and add it to the relevant SessionGroup object
        # in the 'groups_by_name' dict. We create the SessionGroup object, if this
        # is the first session of that group we encounter.
        groups_by_name = {}
        # The TensorBoard runs with session start info are the
        # "sessions", which are not necessarily the runs that actually
        # contain metrics (may be in subdirectories).
        session_names = [
            run
            for (run, tags) in self._hparams_run_to_tag_to_content.items()
            if metadata.SESSION_START_INFO_TAG in tags
        ]
        metric_runs = set()
        metric_tags = set()
        for session_name in session_names:
            for metric in self._experiment.metric_infos:
                metric_name = metric.name
                (run, tag) = metrics.run_tag_from_session_and_metric(
                    session_name, metric_name
                )
                metric_runs.add(run)
                metric_tags.add(tag)
        all_metric_evals = self._context.read_last_scalars(
            self._experiment_id,
            run_tag_filter=provider.RunTagFilter(
                runs=metric_runs, tags=metric_tags
            ),
        )
        for (
            session_name,
            tag_to_content,
        ) in self._hparams_run_to_tag_to_content.items():
            if metadata.SESSION_START_INFO_TAG not in tag_to_content:
                continue
            start_info = metadata.parse_session_start_info_plugin_data(
                tag_to_content[metadata.SESSION_START_INFO_TAG]
            )
            end_info = None
            if metadata.SESSION_END_INFO_TAG in tag_to_content:
                end_info = metadata.parse_session_end_info_plugin_data(
                    tag_to_content[metadata.SESSION_END_INFO_TAG]
                )
            session = self._build_session(
                session_name, start_info, end_info, all_metric_evals
            )
            if session.status in self._request.allowed_statuses:
                self._add_session(session, start_info, groups_by_name)

        # Compute the session group's aggregated metrics for each group.
        groups = groups_by_name.values()
        for group in groups:
            # We sort the sessions in a group so that the order is deterministic.
            group.sessions.sort(key=operator.attrgetter("name"))
            self._aggregate_metrics(group)
        return groups

    def _add_session(self, session, start_info, groups_by_name):
        """Adds a new Session protobuffer to the 'groups_by_name' dictionary.

        Called by _build_session_groups when we encounter a new session. Creates
        the Session protobuffer and adds it to the relevant group in the
        'groups_by_name' dict. Creates the session group if this is the first time
        we encounter it.

        Args:
          session: api_pb2.Session. The session to add.
          start_info: The SessionStartInfo protobuffer associated with the session.
          groups_by_name: A str to SessionGroup protobuffer dict. Representing the
            session groups and sessions found so far.
        """
        # If the group_name is empty, this session's group contains only
        # this session. Use the session name for the group name since session
        # names are unique.
        group_name = start_info.group_name or session.name
        if group_name in groups_by_name:
            groups_by_name[group_name].sessions.extend([session])
        else:
            # Create the group and add the session as the first one.
            group = api_pb2.SessionGroup(
                name=group_name,
                sessions=[session],
                monitor_url=start_info.monitor_url,
            )
            # Copy hparams from the first session (all sessions should have the same
            # hyperparameter values) into result.
            # There doesn't seem to be a way to initialize a protobuffer map in the
            # constructor.
            for (key, value) in six.iteritems(start_info.hparams):
                group.hparams[key].CopyFrom(value)
            groups_by_name[group_name] = group

    def _build_session(self, name, start_info, end_info, all_metric_evals):
        """Builds a session object."""

        assert start_info is not None
        result = api_pb2.Session(
            name=name,
            start_time_secs=start_info.start_time_secs,
            model_uri=start_info.model_uri,
            metric_values=self._build_session_metric_values(
                name, all_metric_evals
            ),
            monitor_url=start_info.monitor_url,
        )
        if end_info is not None:
            result.status = end_info.status
            result.end_time_secs = end_info.end_time_secs
        return result

    def _build_session_metric_values(self, session_name, all_metric_evals):
        """Builds the session metric values."""

        # result is a list of api_pb2.MetricValue instances.
        result = []
        metric_infos = self._experiment.metric_infos
        for metric_info in metric_infos:
            metric_name = metric_info.name
            (run, tag) = metrics.run_tag_from_session_and_metric(
                session_name, metric_name
            )
            datum = all_metric_evals.get(run, {}).get(tag)
            if not datum:
                # It's ok if we don't find the metric in the session.
                # We skip it here. For filtering and sorting purposes its value is None.
                continue
            result.append(
                api_pb2.MetricValue(
                    name=metric_name,
                    wall_time_secs=datum.wall_time,
                    training_step=datum.step,
                    value=datum.value,
                )
            )
        return result

    def _aggregate_metrics(self, session_group):
        """Sets the metrics of the group based on aggregation_type."""

        if (
            self._request.aggregation_type == api_pb2.AGGREGATION_AVG
            or self._request.aggregation_type == api_pb2.AGGREGATION_UNSET
        ):
            _set_avg_session_metrics(session_group)
        elif self._request.aggregation_type == api_pb2.AGGREGATION_MEDIAN:
            _set_median_session_metrics(
                session_group, self._request.aggregation_metric
            )
        elif self._request.aggregation_type == api_pb2.AGGREGATION_MIN:
            _set_extremum_session_metrics(
                session_group, self._request.aggregation_metric, min
            )
        elif self._request.aggregation_type == api_pb2.AGGREGATION_MAX:
            _set_extremum_session_metrics(
                session_group, self._request.aggregation_metric, max
            )
        else:
            raise error.HParamsError(
                "Unknown aggregation_type in request: %s"
                % self._request.aggregation_type
            )

    def _filter(self, session_groups):
        return [sg for sg in session_groups if self._passes_all_filters(sg)]

    def _passes_all_filters(self, session_group):
        return all(filter_fn(session_group) for filter_fn in self._filters)

    def _sort(self, session_groups):
        """Sorts 'session_groups' in place according to _request.col_params."""

        # Sort by session_group name so we have a deterministic order.
        session_groups.sort(key=operator.attrgetter("name"))
        # Sort by lexicographical order of the _request.col_params whose order
        # is not ORDER_UNSPECIFIED. The first such column is the primary sorting
        # key, the second is the secondary sorting key, etc. To achieve that we
        # need to iterate on these columns in reverse order (thus the primary key
        # is the key used in the last sort).
        for col_param, extractor in reversed(
            list(zip(self._request.col_params, self._extractors))
        ):
            if col_param.order == api_pb2.ORDER_UNSPECIFIED:
                continue
            if col_param.order == api_pb2.ORDER_ASC:
                session_groups.sort(
                    key=_create_key_func(
                        extractor,
                        none_is_largest=not col_param.missing_values_first,
                    )
                )
            elif col_param.order == api_pb2.ORDER_DESC:
                session_groups.sort(
                    key=_create_key_func(
                        extractor,
                        none_is_largest=col_param.missing_values_first,
                    ),
                    reverse=True,
                )
            else:
                raise error.HParamsError(
                    "Unknown col_param.order given: %s" % col_param
                )

    def _create_response(self, session_groups):
        return api_pb2.ListSessionGroupsResponse(
            session_groups=session_groups[
                self._request.start_index : self._request.start_index
                + self._request.slice_size
            ],
            total_size=len(session_groups),
        )


def _create_key_func(extractor, none_is_largest):
    """Returns a key_func to be used in list.sort().

    Returns a key_func to be used in list.sort() that sorts session groups
    by the value extracted by extractor. 'None' extracted values will either
    be considered largest or smallest as specified by the "none_is_largest"
    boolean parameter.

    Args:
      extractor: An extractor function that extract the key from the session
        group.
      none_is_largest: bool. If true treats 'None's as largest; otherwise
        smallest.
    """
    if none_is_largest:

        def key_func_none_is_largest(session_group):
            value = extractor(session_group)
            return (value is None, value)

        return key_func_none_is_largest

    def key_func_none_is_smallest(session_group):
        value = extractor(session_group)
        return (value is not None, value)

    return key_func_none_is_smallest


# Extractors. An extractor is a function that extracts some property (a metric
# or a hyperparameter) from a SessionGroup instance.
def _create_extractors(col_params):
    """Creates extractors to extract properties corresponding to 'col_params'.

    Args:
      col_params: List of ListSessionGroupsRequest.ColParam protobufs.
    Returns:
      A list of extractor functions. The ith element in the
      returned list extracts the column corresponding to the ith element of
      _request.col_params
    """
    result = []
    for col_param in col_params:
        result.append(_create_extractor(col_param))
    return result


def _create_extractor(col_param):
    if col_param.HasField("metric"):
        return _create_metric_extractor(col_param.metric)
    elif col_param.HasField("hparam"):
        return _create_hparam_extractor(col_param.hparam)
    else:
        raise error.HParamsError(
            'Got ColParam with both "metric" and "hparam" fields unset: %s'
            % col_param
        )


def _create_metric_extractor(metric_name):
    """Returns function that extracts a metric from a session group or a
    session.

    Args:
      metric_name: tensorboard.hparams.MetricName protobuffer. Identifies the
      metric to extract from the session group.
    Returns:
      A function that takes a tensorboard.hparams.SessionGroup or
      tensorborad.hparams.Session protobuffer and returns the value of the metric
      identified by 'metric_name' or None if the value doesn't exist.
    """

    def extractor_fn(session_or_group):
        metric_value = _find_metric_value(session_or_group, metric_name)
        return metric_value.value if metric_value else None

    return extractor_fn


def _find_metric_value(session_or_group, metric_name):
    """Returns the metric_value for a given metric in a session or session
    group.

    Args:
      session_or_group: A Session protobuffer or SessionGroup protobuffer.
      metric_name: A MetricName protobuffer. The metric to search for.
    Returns:
      A MetricValue protobuffer representing the value of the given metric or
      None if no such metric was found in session_or_group.
    """
    # Note: We can speed this up by converting the metric_values field
    # to a dictionary on initialization, to avoid a linear search here. We'll
    # need to wrap the SessionGroup and Session protos in a python object for
    # that.
    for metric_value in session_or_group.metric_values:
        if (
            metric_value.name.tag == metric_name.tag
            and metric_value.name.group == metric_name.group
        ):
            return metric_value


def _create_hparam_extractor(hparam_name):
    """Returns an extractor function that extracts an hparam from a session
    group.

    Args:
      hparam_name: str. Identies the hparam to extract from the session group.
    Returns:
      A function that takes a tensorboard.hparams.SessionGroup protobuffer and
      returns the value, as a native Python object, of the hparam identified by
      'hparam_name'.
    """

    def extractor_fn(session_group):
        if hparam_name in session_group.hparams:
            return _value_to_python(session_group.hparams[hparam_name])
        return None

    return extractor_fn


# Filters. A filter is a boolean function that takes a session group and returns
# True if it should be included in the result. Currently, Filters are functions
# of a single column value extracted from the session group with a given
# extractor specified in the construction of the filter.
def _create_filters(col_params, extractors):
    """Creates filters for the given col_params.

    Args:
      col_params: List of ListSessionGroupsRequest.ColParam protobufs.
      extractors: list of extractor functions of the same length as col_params.
        Each element should extract the column described by the corresponding
        element of col_params.
    Returns:
      A list of filter functions. Each corresponding to a single
      col_params.filter oneof field of _request
    """
    result = []
    for col_param, extractor in zip(col_params, extractors):
        a_filter = _create_filter(col_param, extractor)
        if a_filter:
            result.append(a_filter)
    return result


def _create_filter(col_param, extractor):
    """Creates a filter for the given col_param and extractor.

    Args:
      col_param: A tensorboard.hparams.ColParams object identifying the column
        and describing the filter to apply.
      extractor: A function that extract the column value identified by
        'col_param' from a tensorboard.hparams.SessionGroup protobuffer.
    Returns:
      A boolean function taking a tensorboard.hparams.SessionGroup protobuffer
      returning True if the session group passes the filter described by
      'col_param'. If col_param does not specify a filter (i.e. any session
      group passes) returns None.
    """
    include_missing_values = not col_param.exclude_missing_values
    if col_param.HasField("filter_regexp"):
        value_filter_fn = _create_regexp_filter(col_param.filter_regexp)
    elif col_param.HasField("filter_interval"):
        value_filter_fn = _create_interval_filter(col_param.filter_interval)
    elif col_param.HasField("filter_discrete"):
        value_filter_fn = _create_discrete_set_filter(col_param.filter_discrete)
    elif include_missing_values:
        # No 'filter' field and include_missing_values is True.
        # Thus, the resulting filter always returns True, so to optimize for this
        # common case we do not include it in the list of filters to check.
        return None
    else:
        value_filter_fn = lambda _: True

    def filter_fn(session_group):
        value = extractor(session_group)
        if value is None:
            return include_missing_values
        return value_filter_fn(value)

    return filter_fn


def _create_regexp_filter(regex):
    """Returns a boolean function that filters strings based on a regular exp.

    Args:
      regex: A string describing the regexp to use.
    Returns:
      A function taking a string and returns True if any of its substrings
      matches regex.
    """
    # Warning: Note that python's regex library allows inputs that take
    # exponential time. Time-limiting it is difficult. When we move to
    # a true multi-tenant tensorboard server, the regexp implementation here
    # would need to be replaced by something more secure.
    compiled_regex = re.compile(regex)

    def filter_fn(value):
        if not isinstance(value, six.string_types):
            raise error.HParamsError(
                "Cannot use a regexp filter for a value of type %s. Value: %s"
                % (type(value), value)
            )
        return re.search(compiled_regex, value) is not None

    return filter_fn


def _create_interval_filter(interval):
    """Returns a function that checkes whether a number belongs to an interval.

    Args:
      interval: A tensorboard.hparams.Interval protobuf describing the interval.
    Returns:
      A function taking a number (a float or an object of a type in
      six.integer_types) that returns True if the number belongs to (the closed)
      'interval'.
    """

    def filter_fn(value):
        if not isinstance(value, six.integer_types) and not isinstance(
            value, float
        ):
            raise error.HParamsError(
                "Cannot use an interval filter for a value of type: %s, Value: %s"
                % (type(value), value)
            )
        return interval.min_value <= value and value <= interval.max_value

    return filter_fn


def _create_discrete_set_filter(discrete_set):
    """Returns a function that checks whether a value belongs to a set.

    Args:
      discrete_set: A list of objects representing the set.
    Returns:
      A function taking an object and returns True if its in the set. Membership
      is tested using the Python 'in' operator (thus, equality of distinct
      objects is computed using the '==' operator).
    """

    def filter_fn(value):
        return value in discrete_set

    return filter_fn


def _value_to_python(value):
    """Converts a google.protobuf.Value to a native Python object."""

    assert isinstance(value, struct_pb2.Value)
    field = value.WhichOneof("kind")
    if field == "number_value":
        return value.number_value
    elif field == "string_value":
        return value.string_value
    elif field == "bool_value":
        return value.bool_value
    else:
        raise ValueError("Unknown struct_pb2.Value oneof field set: %s" % field)


# As protobuffers are mutable we can't use MetricName directly as a dict's key.
# Instead, we duplicate MetricName as a collections.namedtuple. Another
# alternative would be to wrap the MetricName protocol buffer class in an
# immutable class that defines equality and __hash__ methods based on the text
# representation of the protocol buffer. This is more complex, but won't
# require modification if we ever add fields to MetricName.
_MetricIdentifier = collections.namedtuple("_MetricIdentifier", "group tag")


class _MetricStats(object):
    """A simple class to hold metric stats used in calculating metric averages.

    Used in _set_avg_session_metrics(). See the comments in that function
    for more details.

    Attributes:
      total: int. The sum of the metric measurements seen so far.
      count: int. The number of largest-step measuremens seen so far.
      total_step: int. The sum of the steps at which the measurements were taken
      total_wall_time_secs: float. The sum of the wall_time_secs at
          which the measurements were taken.
    """

    # We use slots here to catch typos in attributes earlier. Note that this makes
    # this class incompatible with 'pickle'.
    __slots__ = [
        "total",
        "count",
        "total_step",
        "total_wall_time_secs",
    ]

    def __init__(self):
        self.total = 0
        self.count = 0
        self.total_step = 0
        self.total_wall_time_secs = 0.0


def _set_avg_session_metrics(session_group):
    """Sets the metrics for the group to be the average of its sessions.

    The resulting session group metrics consist of the union of metrics across
    the group's sessions. The value of each session group metric is the average
    of that metric values across the sessions in the group. The 'step' and
    'wall_time_secs' fields of the resulting MetricValue field in the session
    group are populated with the corresponding averages (truncated for 'step')
    as well.

    Args:
      session_group: A SessionGroup protobuffer.
    """
    assert session_group.sessions, "SessionGroup cannot be empty."
    # Algorithm: Iterate over all (session, metric) pairs and maintain a
    # dict from _MetricIdentifier to _MetricStats objects.
    # Then use the final dict state to compute the average for each metric.
    metric_stats = collections.defaultdict(_MetricStats)
    for session in session_group.sessions:
        for metric_value in session.metric_values:
            metric_name = _MetricIdentifier(
                group=metric_value.name.group, tag=metric_value.name.tag
            )
            stats = metric_stats[metric_name]
            stats.total += metric_value.value
            stats.count += 1
            stats.total_step += metric_value.training_step
            stats.total_wall_time_secs += metric_value.wall_time_secs

    del session_group.metric_values[:]
    for (metric_name, stats) in six.iteritems(metric_stats):
        session_group.metric_values.add(
            name=api_pb2.MetricName(
                group=metric_name.group, tag=metric_name.tag
            ),
            value=float(stats.total) / float(stats.count),
            training_step=stats.total_step // stats.count,
            wall_time_secs=stats.total_wall_time_secs / stats.count,
        )


# A namedtuple to hold a session's metric value.
# 'session_index' is the index of the session in its group.
_Measurement = collections.namedtuple(
    "_Measurement", ["metric_value", "session_index"]
)


def _set_median_session_metrics(session_group, aggregation_metric):
    """Sets the metrics for session_group to those of its "median session".

    The median session is the session in session_group with the median value
    of the metric given by 'aggregation_metric'. The median is taken over the
    subset of sessions in the group whose 'aggregation_metric' was measured
    at the largest training step among the sessions in the group.

    Args:
      session_group: A SessionGroup protobuffer.
      aggregation_metric: A MetricName protobuffer.
    """
    measurements = sorted(
        _measurements(session_group, aggregation_metric),
        key=operator.attrgetter("metric_value.value"),
    )
    median_session = measurements[(len(measurements) - 1) // 2].session_index
    del session_group.metric_values[:]
    session_group.metric_values.MergeFrom(
        session_group.sessions[median_session].metric_values
    )


def _set_extremum_session_metrics(
    session_group, aggregation_metric, extremum_fn
):
    """Sets the metrics for session_group to those of its "extremum session".

    The extremum session is the session in session_group with the extremum value
    of the metric given by 'aggregation_metric'. The extremum is taken over the
    subset of sessions in the group whose 'aggregation_metric' was measured
    at the largest training step among the sessions in the group.

    Args:
      session_group: A SessionGroup protobuffer.
      aggregation_metric: A MetricName protobuffer.
      extremum_fn: callable. Must be either 'min' or 'max'. Determines the type of
        extremum to compute.
    """
    measurements = _measurements(session_group, aggregation_metric)
    ext_session = extremum_fn(
        measurements, key=operator.attrgetter("metric_value.value")
    ).session_index
    del session_group.metric_values[:]
    session_group.metric_values.MergeFrom(
        session_group.sessions[ext_session].metric_values
    )


def _measurements(session_group, metric_name):
    """A generator for the values of the metric across the sessions in the
    group.

    Args:
      session_group: A SessionGroup protobuffer.
      metric_name: A MetricName protobuffer.
    Yields:
      The next metric value wrapped in a _Measurement instance.
    """
    for session_index, session in enumerate(session_group.sessions):
        metric_value = _find_metric_value(session, metric_name)
        if not metric_value:
            continue
        yield _Measurement(metric_value, session_index)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities to migrate legacy protos to their modern equivalents."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from tensorboard.compat import tf
from tensorboard.compat.proto import event_pb2
from tensorboard.compat.proto import summary_pb2
from tensorboard.plugins.audio import metadata as audio_metadata
from tensorboard.plugins.histogram import metadata as histogram_metadata
from tensorboard.plugins.image import metadata as image_metadata
from tensorboard.plugins.scalar import metadata as scalar_metadata
from tensorboard.util import tensor_util


def migrate_event(event):
    if not event.HasField("summary"):
        return event
    old_values = event.summary.value
    new_values = [migrate_value(value) for value in old_values]
    # Optimization: Don't create a new event if there were no changes.
    if len(old_values) == len(new_values) and all(
        x is y for (x, y) in zip(old_values, new_values)
    ):
        return event
    result = event_pb2.Event()
    result.CopyFrom(event)
    del result.summary.value[:]
    result.summary.value.extend(new_values)
    return result


def migrate_value(value):
    """Convert `value` to a new-style value, if necessary and possible.

    An "old-style" value is a value that uses any `value` field other than
    the `tensor` field. A "new-style" value is a value that uses the
    `tensor` field. TensorBoard continues to support old-style values on
    disk; this method converts them to new-style values so that further
    code need only deal with one data format.

    Arguments:
      value: A `Summary.Value` object. This argument is not modified.

    Returns:
      If the `value` is an old-style value for which there is a new-style
      equivalent, the result is the new-style value. Otherwise---if the
      value is already new-style or does not yet have a new-style
      equivalent---the value will be returned unchanged.

    :type value: Summary.Value
    :rtype: Summary.Value
    """
    handler = {
        "histo": _migrate_histogram_value,
        "image": _migrate_image_value,
        "audio": _migrate_audio_value,
        "simple_value": _migrate_scalar_value,
    }.get(value.WhichOneof("value"))
    return handler(value) if handler else value


def make_summary(tag, metadata, data):
    tensor_proto = tensor_util.make_tensor_proto(data)
    return summary_pb2.Summary.Value(
        tag=tag, metadata=metadata, tensor=tensor_proto
    )


def _migrate_histogram_value(value):
    histogram_value = value.histo
    bucket_lefts = [histogram_value.min] + histogram_value.bucket_limit[:-1]
    bucket_rights = histogram_value.bucket_limit[:-1] + [histogram_value.max]
    bucket_counts = histogram_value.bucket
    buckets = np.array(
        [bucket_lefts, bucket_rights, bucket_counts], dtype=np.float32
    ).transpose()

    summary_metadata = histogram_metadata.create_summary_metadata(
        display_name=value.metadata.display_name or value.tag,
        description=value.metadata.summary_description,
    )

    return make_summary(value.tag, summary_metadata, buckets)


def _migrate_image_value(value):
    image_value = value.image
    data = [
        tf.compat.as_bytes(str(image_value.width)),
        tf.compat.as_bytes(str(image_value.height)),
        tf.compat.as_bytes(image_value.encoded_image_string),
    ]

    summary_metadata = image_metadata.create_summary_metadata(
        display_name=value.metadata.display_name or value.tag,
        description=value.metadata.summary_description,
    )
    return make_summary(value.tag, summary_metadata, data)


def _migrate_audio_value(value):
    audio_value = value.audio
    data = [[audio_value.encoded_audio_string, b""]]  # empty label
    summary_metadata = audio_metadata.create_summary_metadata(
        display_name=value.metadata.display_name or value.tag,
        description=value.metadata.summary_description,
        encoding=audio_metadata.Encoding.Value("WAV"),
    )
    return make_summary(value.tag, summary_metadata, data)


def _migrate_scalar_value(value):
    scalar_value = value.simple_value
    summary_metadata = scalar_metadata.create_summary_metadata(
        display_name=value.metadata.display_name or value.tag,
        description=value.metadata.summary_description,
    )
    return make_summary(value.tag, summary_metadata, scalar_value)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Manages writing debugger-related events to disk.

Creates new events files if previous files get too big.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import re
import threading
import time

import tensorflow as tf

# To keep compatibility with both 1.x and 2.x
try:
    from tensorflow.python import _pywrap_events_writer as tf_events_writer
except ImportError:
    from tensorflow.python import pywrap_tensorflow as tf_events_writer
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

# Files containing debugger-related events should start with this string.
DEBUGGER_EVENTS_FILE_STARTING_TEXT = "events.debugger"

# A regex used to match the names of debugger-related events files.
_DEBUGGER_EVENTS_FILE_NAME_REGEX = re.compile(
    r"^" + re.escape(DEBUGGER_EVENTS_FILE_STARTING_TEXT) + r"\.\d+\.(\d+)"
)

# The default events file size cap (in bytes).
_DEFAULT_EVENTS_FILE_SIZE_CAP_BYTES = 500 * 1024 * 1024

# By default, we check whether the current events file has exceeded the file
# size cap every this many steps.
_DEFAULT_CHECK_EVENT_FILES_SIZE_CAP_EVERY = 1e5

# When the total size that all events on disk exceeds this value, the events
# writer manager deletes least recently created debugger-related events files.
_DEFAULT_TOTAL_SIZE_CAP_BYTES = 1e9


class EventsWriterManager(object):
    """Manages writing debugger-related events to disk.

    Creates new events writers if files get too big.
    """

    def __init__(
        self,
        events_directory,
        single_file_size_cap_bytes=_DEFAULT_EVENTS_FILE_SIZE_CAP_BYTES,
        check_this_often=_DEFAULT_CHECK_EVENT_FILES_SIZE_CAP_EVERY,
        total_file_size_cap_bytes=_DEFAULT_TOTAL_SIZE_CAP_BYTES,
        always_flush=False,
    ):
        """Constructs an EventsWriterManager.

        Args:
          events_directory: (`string`) The log directory in which debugger events
            reside.
          single_file_size_cap_bytes: (`int`) A number of bytes. During a check, if
            the manager determines that the events file being written to exceeds
            this size, it creates a new events file to write to. Note that events
            file may still exceed this size - the events writer manager just creates
            a new events file if it finds that the current file exceeds this size.
          check_this_often: (`int`) The manager performs a file size check every
            this many events. We want to avoid checking upon every event for
            performance reasons. If provided, must be greater than 1.
          total_file_size_cap_bytes: A cap on the total number of bytes occupied by
            all events. When a new events writer is created, the least recently
            created events file will be deleted if the total size occupied by
            debugger-related events on disk exceeds this cap. Note that the total
            size could now and then be larger than this cap because the events
            writer manager only checks when it creates a new events file.
          always_flush: (`bool`) Whether to flush to disk after every write. Useful
            for testing.
        """
        self._events_directory = events_directory
        self._single_file_size_cap_bytes = single_file_size_cap_bytes
        self.total_file_size_cap_bytes = total_file_size_cap_bytes
        self._check_this_often = check_this_often
        self._always_flush = always_flush

        # Each events file gets a unique file count within its file name. This value
        # increments every time a new events file is created.
        self._events_file_count = 0

        # If there are existing event files, assign the events file count to be
        # greater than the last existing one.
        events_file_names = self._fetch_events_files_on_disk()
        if events_file_names:
            self._events_file_count = (
                self._obtain_file_index(events_file_names[-1]) + 1
            )

        self._event_count = 0
        self._lock = threading.Lock()
        self._events_writer = self._create_events_writer(events_directory)

    def write_event(self, event):
        """Writes an event proto to disk.

        This method is threadsafe with respect to invocations of itself.

        Args:
          event: The event proto.

        Raises:
          IOError: If writing the event proto to disk fails.
        """
        self._lock.acquire()
        try:
            self._events_writer.WriteEvent(event)
            self._event_count += 1
            if self._always_flush:
                # We flush on every event within the integration test.
                self._events_writer.Flush()

            if self._event_count == self._check_this_often:
                # Every so often, we check whether the size of the file is too big.
                self._event_count = 0

                # Flush to get an accurate size check.
                self._events_writer.Flush()

                file_path = os.path.join(
                    self._events_directory, self.get_current_file_name()
                )
                if not tf.io.gfile.exists(file_path):
                    # The events file does not exist. Perhaps the user had manually
                    # deleted it after training began. Create a new one.
                    self._events_writer.Close()
                    self._events_writer = self._create_events_writer(
                        self._events_directory
                    )
                elif (
                    tf.io.gfile.stat(file_path).length
                    > self._single_file_size_cap_bytes
                ):
                    # The current events file has gotten too big. Close the previous
                    # events writer. Make a new one.
                    self._events_writer.Close()
                    self._events_writer = self._create_events_writer(
                        self._events_directory
                    )
        except IOError as err:
            logger.error(
                "Writing to %s failed: %s", self.get_current_file_name(), err
            )
        self._lock.release()

    def get_current_file_name(self):
        """Gets the name of the events file currently being written to.

        Returns:
          The name of the events file being written to.
        """
        return tf.compat.as_text(self._events_writer.FileName())

    def dispose(self):
        """Disposes of this events writer manager, making it no longer usable.

        Call this method when this object is done being used in order to
        clean up resources and handlers. This method should ever only be
        called once.
        """
        self._lock.acquire()
        self._events_writer.Close()
        self._events_writer = None
        self._lock.release()

    def _create_events_writer(self, directory):
        """Creates a new events writer.

        Args:
          directory: The directory in which to write files containing events.

        Returns:
          A new events writer, which corresponds to a new events file.
        """
        total_size = 0
        events_files = self._fetch_events_files_on_disk()
        for file_name in events_files:
            file_path = os.path.join(self._events_directory, file_name)
            total_size += tf.io.gfile.stat(file_path).length

        if total_size >= self.total_file_size_cap_bytes:
            # The total size written to disk is too big. Delete events files until
            # the size is below the cap.
            for file_name in events_files:
                if total_size < self.total_file_size_cap_bytes:
                    break

                file_path = os.path.join(self._events_directory, file_name)
                file_size = tf.io.gfile.stat(file_path).length
                try:
                    tf.io.gfile.remove(file_path)
                    total_size -= file_size
                    logger.info(
                        "Deleted %s because events files take up over %d bytes",
                        file_path,
                        self.total_file_size_cap_bytes,
                    )
                except IOError as err:
                    logger.error("Deleting %s failed: %s", file_path, err)

        # We increment this index because each events writer must differ in prefix.
        self._events_file_count += 1
        file_path = "%s.%d.%d" % (
            os.path.join(directory, DEBUGGER_EVENTS_FILE_STARTING_TEXT),
            time.time(),
            self._events_file_count,
        )
        logger.info("Creating events file %s", file_path)
        return tf_events_writer.EventsWriter(tf.compat.as_bytes(file_path))

    def _fetch_events_files_on_disk(self):
        """Obtains the names of debugger-related events files within the
        directory.

        Returns:
          The names of the debugger-related events files written to disk. The names
          are sorted in increasing events file index.
        """
        all_files = tf.io.gfile.listdir(self._events_directory)
        relevant_files = [
            file_name
            for file_name in all_files
            if _DEBUGGER_EVENTS_FILE_NAME_REGEX.match(file_name)
        ]
        return sorted(relevant_files, key=self._obtain_file_index)

    def _obtain_file_index(self, file_name):
        """Obtains the file index associated with an events file.

        The index is stored within a file name and is incremented every time a new
        events file is created. Assumes that the file name is a valid debugger
        events file name.

        Args:
          file_name: The name of the debugger-related events file. The file index is
              stored within the file name.

        Returns:
          The integer events file index.
        """
        return int(_DEBUGGER_EVENTS_FILE_NAME_REGEX.match(file_name).group(1))



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===========================================================================
"""A loading-only EventMultiplexer that actually populates a SQLite DB."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import collections
import os
import threading
import time

import six
from six.moves import queue, xrange  # pylint: disable=redefined-builtin

from tensorboard import data_compat
from tensorboard.backend.event_processing import directory_watcher
from tensorboard.backend.event_processing import event_file_loader
from tensorboard.backend.event_processing import io_wrapper
from tensorboard.backend.event_processing import plugin_event_multiplexer
from tensorboard.backend.event_processing import sqlite_writer
from tensorboard.compat import tf
from tensorboard.compat.proto import event_pb2
from tensorboard.util import tb_logging


logger = tb_logging.get_logger()


class DbImportMultiplexer(plugin_event_multiplexer.EventMultiplexer):
    """A loading-only `EventMultiplexer` that populates a SQLite DB.

    This EventMultiplexer only loads data; the read APIs always return
    empty results, since all data is accessed instead via SQL against
    the db_connection_provider wrapped by this multiplexer.
    """

    def __init__(
        self,
        db_uri,
        db_connection_provider,
        purge_orphaned_data,
        max_reload_threads,
    ):
        """Constructor for `DbImportMultiplexer`.

        Args:
          db_uri: A URI to the database file in use.
          db_connection_provider: Provider function for creating a DB connection.
          purge_orphaned_data: Whether to discard any events that were "orphaned" by
            a TensorFlow restart.
          max_reload_threads: The max number of threads that TensorBoard can use
            to reload runs. Each thread reloads one run at a time. If not provided,
            reloads runs serially (one after another).
        """
        logger.info("DbImportMultiplexer initializing for %s", db_uri)
        super(DbImportMultiplexer, self).__init__()
        self.db_uri = db_uri
        self.db_connection_provider = db_connection_provider
        self._purge_orphaned_data = purge_orphaned_data
        self._max_reload_threads = max_reload_threads
        self._event_sink = None
        self._run_loaders = {}

        if self._purge_orphaned_data:
            logger.warn(
                "--db_import does not yet support purging orphaned data"
            )

        conn = self.db_connection_provider()
        # Set the DB in WAL mode so reads don't block writes.
        conn.execute("PRAGMA journal_mode=wal")
        conn.execute("PRAGMA synchronous=normal")  # Recommended for WAL mode
        sqlite_writer.initialize_schema(conn)
        logger.info("DbImportMultiplexer done initializing")

    def AddRun(self, path, name=None):
        """Unsupported; instead use AddRunsFromDirectory."""
        raise NotImplementedError("Unsupported; use AddRunsFromDirectory()")

    def AddRunsFromDirectory(self, path, name=None):
        """Load runs from a directory; recursively walks subdirectories.

        If path doesn't exist, no-op. This ensures that it is safe to call
          `AddRunsFromDirectory` multiple times, even before the directory is made.

        Args:
          path: A string path to a directory to load runs from.
          name: Optional, specifies a name for the experiment under which the
            runs from this directory hierarchy will be imported. If omitted, the
            path will be used as the name.

        Raises:
          ValueError: If the path exists and isn't a directory.
        """
        logger.info("Starting AddRunsFromDirectory: %s (as %s)", path, name)
        for subdir in io_wrapper.GetLogdirSubdirectories(path):
            logger.info("Processing directory %s", subdir)
            if subdir not in self._run_loaders:
                logger.info("Creating DB loader for directory %s", subdir)
                names = self._get_exp_and_run_names(path, subdir, name)
                experiment_name, run_name = names
                self._run_loaders[subdir] = _RunLoader(
                    subdir=subdir,
                    experiment_name=experiment_name,
                    run_name=run_name,
                )
        logger.info("Done with AddRunsFromDirectory: %s", path)

    def Reload(self):
        """Load events from every detected run."""
        logger.info("Beginning DbImportMultiplexer.Reload()")
        # Defer event sink creation until needed; this ensures it will only exist in
        # the thread that calls Reload(), since DB connections must be thread-local.
        if not self._event_sink:
            self._event_sink = _SqliteWriterEventSink(
                self.db_connection_provider
            )
        # Use collections.deque() for speed when we don't need blocking since it
        # also has thread-safe appends/pops.
        loader_queue = collections.deque(six.itervalues(self._run_loaders))
        loader_delete_queue = collections.deque()

        def batch_generator():
            while True:
                try:
                    loader = loader_queue.popleft()
                except IndexError:
                    return
                try:
                    for batch in loader.load_batches():
                        yield batch
                except directory_watcher.DirectoryDeletedError:
                    loader_delete_queue.append(loader)
                except (OSError, IOError) as e:
                    logger.error("Unable to load run %r: %s", loader.subdir, e)

        num_threads = min(self._max_reload_threads, len(self._run_loaders))
        if num_threads <= 1:
            logger.info("Importing runs serially on a single thread")
            for batch in batch_generator():
                self._event_sink.write_batch(batch)
        else:
            output_queue = queue.Queue()
            sentinel = object()

            def producer():
                try:
                    for batch in batch_generator():
                        output_queue.put(batch)
                finally:
                    output_queue.put(sentinel)

            logger.info("Starting %d threads to import runs", num_threads)
            for i in xrange(num_threads):
                thread = threading.Thread(target=producer, name="Loader %d" % i)
                thread.daemon = True
                thread.start()
            num_live_threads = num_threads
            while num_live_threads > 0:
                output = output_queue.get()
                if output == sentinel:
                    num_live_threads -= 1
                    continue
                self._event_sink.write_batch(output)
        for loader in loader_delete_queue:
            logger.warn("Deleting loader %r", loader.subdir)
            del self._run_loaders[loader.subdir]
        logger.info("Finished with DbImportMultiplexer.Reload()")

    def _get_exp_and_run_names(
        self, path, subdir, experiment_name_override=None
    ):
        if experiment_name_override is not None:
            return (experiment_name_override, os.path.relpath(subdir, path))
        sep = io_wrapper.PathSeparator(path)
        path_parts = os.path.relpath(subdir, path).split(sep, 1)
        experiment_name = path_parts[0]
        run_name = path_parts[1] if len(path_parts) == 2 else "."
        return (experiment_name, run_name)


# Struct holding a list of tf.Event serialized protos along with metadata about
# the associated experiment and run.
_EventBatch = collections.namedtuple(
    "EventBatch", ["events", "experiment_name", "run_name"]
)


class _RunLoader(object):
    """Loads a single run directory in batches."""

    _BATCH_COUNT = 5000
    _BATCH_BYTES = 2 ** 20  # 1 MiB

    def __init__(self, subdir, experiment_name, run_name):
        """Constructs a `_RunLoader`.

        Args:
          subdir: string, filesystem path of the run directory
          experiment_name: string, name of the run's experiment
          run_name: string, name of the run
        """
        self._subdir = subdir
        self._experiment_name = experiment_name
        self._run_name = run_name
        self._directory_watcher = directory_watcher.DirectoryWatcher(
            subdir,
            event_file_loader.RawEventFileLoader,
            io_wrapper.IsTensorFlowEventsFile,
        )

    @property
    def subdir(self):
        return self._subdir

    def load_batches(self):
        """Returns a batched event iterator over the run directory event
        files."""
        event_iterator = self._directory_watcher.Load()
        while True:
            events = []
            event_bytes = 0
            start = time.time()
            for event_proto in event_iterator:
                events.append(event_proto)
                event_bytes += len(event_proto)
                if (
                    len(events) >= self._BATCH_COUNT
                    or event_bytes >= self._BATCH_BYTES
                ):
                    break
            elapsed = time.time() - start
            logger.debug(
                "RunLoader.load_batch() yielded in %0.3f sec for %s",
                elapsed,
                self._subdir,
            )
            if not events:
                return
            yield _EventBatch(
                events=events,
                experiment_name=self._experiment_name,
                run_name=self._run_name,
            )


@six.add_metaclass(abc.ABCMeta)
class _EventSink(object):
    """Abstract sink for batches of serialized tf.Event data."""

    @abc.abstractmethod
    def write_batch(self, event_batch):
        """Writes the given event batch to the sink.

        Args:
          event_batch: an _EventBatch of event data.
        """
        raise NotImplementedError()


class _SqliteWriterEventSink(_EventSink):
    """Implementation of EventSink using SqliteWriter."""

    def __init__(self, db_connection_provider):
        """Constructs a SqliteWriterEventSink.

        Args:
          db_connection_provider: Provider function for creating a DB connection.
        """
        self._writer = sqlite_writer.SqliteWriter(db_connection_provider)

    def write_batch(self, event_batch):
        start = time.time()
        tagged_data = {}
        for event_proto in event_batch.events:
            event = event_pb2.Event.FromString(event_proto)
            self._process_event(event, tagged_data)
        if tagged_data:
            self._writer.write_summaries(
                tagged_data,
                experiment_name=event_batch.experiment_name,
                run_name=event_batch.run_name,
            )
        elapsed = time.time() - start
        logger.debug(
            "SqliteWriterEventSink.WriteBatch() took %0.3f sec for %s events",
            elapsed,
            len(event_batch.events),
        )

    def _process_event(self, event, tagged_data):
        """Processes a single tf.Event and records it in tagged_data."""
        event_type = event.WhichOneof("what")
        # Handle the most common case first.
        if event_type == "summary":
            for value in event.summary.value:
                value = data_compat.migrate_value(value)
                tag, metadata, values = tagged_data.get(
                    value.tag, (None, None, [])
                )
                values.append((event.step, event.wall_time, value.tensor))
                if tag is None:
                    # Store metadata only from the first event.
                    tagged_data[value.tag] = sqlite_writer.TagData(
                        value.tag, value.metadata, values
                    )
        elif event_type == "file_version":
            pass  # TODO: reject file version < 2 (at loader level)
        elif event_type == "session_log":
            if event.session_log.status == event_pb2.SessionLog.START:
                pass  # TODO: implement purging via sqlite writer truncation method
        elif event_type in ("graph_def", "meta_graph_def"):
            pass  # TODO: support graphs
        elif event_type == "tagged_run_metadata":
            pass  # TODO: support run metadata



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Data structures and algorithms for numerics alert.

The alerts are generated when a Tensor's elements contain bad values,
including nan, -inf and +inf.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import json
import re

import tensorflow as tf

from tensorflow.python import debug as tf_debug
from tensorboard.plugins.debugger import constants

# The following two namedtuples are the same except that
# 1) `timestamp` in `NumericsAlert` is the timestamp of a single alerting event,
#    while `first_timestamp` in `NumericsAlertReportRow` is the first (earliest)
#    timestamp of a se tof aggregated `NumericsAlert`s.
# 2) The counts in `NumericsAlert` are the counts of elements in the tensor,
#    while the event counts in `NumericsAlertReportRow` are counts of previous
#    `NumericsAlert` events of the corresponding categories.
NumericsAlert = collections.namedtuple(
    "NumericsAlert",
    [
        "device_name",
        "tensor_name",
        "timestamp",
        "nan_count",
        "neg_inf_count",
        "pos_inf_count",
    ],
)

NumericsAlertReportRow = collections.namedtuple(
    "NumericsAlertReportRow",
    [
        "device_name",
        "tensor_name",
        "first_timestamp",
        "nan_event_count",
        "neg_inf_event_count",
        "pos_inf_event_count",
    ],
)

# Used to reconstruct an _EventTracker from data read from disk. When updating
# this named tuple, make sure to keep the properties of _EventTracker in sync.
EventTrackerDescription = collections.namedtuple(
    "EventTrackerDescription",
    ["event_count", "first_timestamp", "last_timestamp"],
)

# Used to reconstruct NumericsAlertHistory.
HistoryTriplet = collections.namedtuple(
    "HistoryTriplet", ["device", "tensor", "jsonable_history"]
)


class _EventTracker(object):
    """Track events for a single category of values (NaN, -Inf, or +Inf)."""

    def __init__(self, event_count=0, first_timestamp=-1, last_timestamp=-1):
        """Tracks events for a single category of values.

        Args:
          event_count: The initial event count to use.
          first_timestamp: The timestamp of the first event with this value.
          last_timestamp: The timestamp of the last event with this category of
              values.
        """

        # When updating the properties of this class, make sure to keep
        # EventTrackerDescription in sync so that data can be written to and from
        # disk correctly.
        self.event_count = event_count
        self.first_timestamp = first_timestamp
        self.last_timestamp = last_timestamp

    def add(self, timestamp):
        if self.event_count == 0:
            self.first_timestamp = timestamp
            self.last_timestamp = timestamp
        else:
            if timestamp < self.first_timestamp:
                self.first_timestamp = timestamp
            if timestamp > self.last_timestamp:
                self.last_timestamp = timestamp
        self.event_count += 1

    def get_description(self):
        return EventTrackerDescription(
            self.event_count, self.first_timestamp, self.last_timestamp
        )


class NumericsAlertHistory(object):
    """History of numerics alerts."""

    def __init__(self, initialization_list=None):
        """Stores alert history for a single device, tensor pair.

        Args:
          initialization_list: (`list`) An optional list parsed from JSON read
            from disk. That entity is used to initialize this NumericsAlertHistory.
            Use the create_jsonable_object method of this class to create such an
            object.
        """
        if initialization_list:
            # Use data to initialize this NumericsAlertHistory.
            self._trackers = {}
            for (
                value_category_key,
                description_list,
            ) in initialization_list.items():
                description = EventTrackerDescription._make(description_list)
                self._trackers[value_category_key] = _EventTracker(
                    event_count=description.event_count,
                    first_timestamp=description.first_timestamp,
                    last_timestamp=description.last_timestamp,
                )
        else:
            # Start cleanly. With no prior data.
            self._trackers = {
                constants.NAN_KEY: _EventTracker(),
                constants.NEG_INF_KEY: _EventTracker(),
                constants.POS_INF_KEY: _EventTracker(),
            }

    def add(self, numerics_alert):
        if numerics_alert.nan_count:
            self._trackers[constants.NAN_KEY].add(numerics_alert.timestamp)
        if numerics_alert.neg_inf_count:
            self._trackers[constants.NEG_INF_KEY].add(numerics_alert.timestamp)
        if numerics_alert.pos_inf_count:
            self._trackers[constants.POS_INF_KEY].add(numerics_alert.timestamp)

    def first_timestamp(self, event_key=None):
        """Obtain the first timestamp.

        Args:
          event_key: the type key of the sought events (e.g., constants.NAN_KEY).
          If None, includes all event type keys.

        Returns:
          First (earliest) timestamp of all the events of the given type (or all
            event types if event_key is None).
        """
        if event_key is None:
            timestamps = [
                self._trackers[key].first_timestamp for key in self._trackers
            ]
            return min(timestamp for timestamp in timestamps if timestamp >= 0)
        else:
            return self._trackers[event_key].first_timestamp

    def last_timestamp(self, event_key=None):
        """Obtain the last timestamp.

        Args:
          event_key: the type key of the sought events (e.g., constants.NAN_KEY). If
          None, includes all event type keys.

        Returns:
          Last (latest) timestamp of all the events of the given type (or all
            event types if event_key is None).
        """
        if event_key is None:
            timestamps = [
                self._trackers[key].first_timestamp for key in self._trackers
            ]
            return max(timestamp for timestamp in timestamps if timestamp >= 0)
        else:
            return self._trackers[event_key].last_timestamp

    def event_count(self, event_key):
        """Obtain event count.

        Args:
          event_key: the type key of the sought events (e.g., constants.NAN_KEY). If
          None, includes all event type keys.

        Returns:
          If event_key is None, return the sum of the event_count of all event
          types. Otherwise, return the event_count of the specified event type.
        """
        return self._trackers[event_key].event_count

    def create_jsonable_history(self):
        """Creates a JSON-able representation of this object.

        Returns:
          A dictionary mapping key to EventTrackerDescription (which can be used to
          create event trackers).
        """
        return {
            value_category_key: tracker.get_description()
            for (value_category_key, tracker) in self._trackers.items()
        }


class NumericsAlertRegistry(object):
    """A registry for alerts on numerics (e.g., due to NaNs and infinities)."""

    def __init__(self, capacity=100, initialization_list=None):
        """Constructor.

        Args:
          capacity: (`int`) maximum number of device-tensor keys to store.
          initialization_list: (`list`) An optional list (parsed from JSON) that
            is used to initialize the data within this registry. Use the
            create_jsonable_registry method of NumericsAlertRegistry to create such
            a list.
        """
        self._capacity = capacity

        # A map from device-tensor key to a the TensorAlertRecord namedtuple.
        # The device-tensor key is a 2-tuple of the format (device_name, node_name).
        # E.g., ("/job:worker/replica:0/task:1/gpu:0", "cross_entropy/Log:0").
        self._data = dict()

        if initialization_list:
            # Initialize the alert registry using the data passed in. This might be
            # backup data used to restore the registry after say a borg pre-emption.
            for entry in initialization_list:
                triplet = HistoryTriplet._make(entry)
                self._data[
                    (triplet.device, triplet.tensor)
                ] = NumericsAlertHistory(
                    initialization_list=triplet.jsonable_history
                )

    def register(self, numerics_alert):
        """Register an alerting numeric event.

        Args:
          numerics_alert: An instance of `NumericsAlert`.
        """
        key = (numerics_alert.device_name, numerics_alert.tensor_name)
        if key in self._data:
            self._data[key].add(numerics_alert)
        else:
            if len(self._data) < self._capacity:
                history = NumericsAlertHistory()
                history.add(numerics_alert)
                self._data[key] = history

    def report(self, device_name_filter=None, tensor_name_filter=None):
        """Get a report of offending device/tensor names.

        The report includes information about the device name, tensor name, first
        (earliest) timestamp of the alerting events from the tensor, in addition to
        counts of nan, positive inf and negative inf events.

        Args:
          device_name_filter: regex filter for device name, or None (not filtered).
          tensor_name_filter: regex filter for tensor name, or None (not filtered).

        Returns:
          A list of NumericsAlertReportRow, sorted by the first_timestamp in
          asecnding order.
        """
        report = []
        for key in self._data:
            device_name, tensor_name = key
            history = self._data[key]
            report.append(
                NumericsAlertReportRow(
                    device_name=device_name,
                    tensor_name=tensor_name,
                    first_timestamp=history.first_timestamp(),
                    nan_event_count=history.event_count(constants.NAN_KEY),
                    neg_inf_event_count=history.event_count(
                        constants.NEG_INF_KEY
                    ),
                    pos_inf_event_count=history.event_count(
                        constants.POS_INF_KEY
                    ),
                )
            )

        if device_name_filter:
            device_name_pattern = re.compile(device_name_filter)
            report = [
                item
                for item in report
                if device_name_pattern.match(item.device_name)
            ]
        if tensor_name_filter:
            tensor_name_pattern = re.compile(tensor_name_filter)
            report = [
                item
                for item in report
                if tensor_name_pattern.match(item.tensor_name)
            ]
        # Sort results chronologically.
        return sorted(report, key=lambda x: x.first_timestamp)

    def create_jsonable_registry(self):
        """Creates a JSON-able representation of this object.

        Returns:
          A dictionary mapping (device, tensor name) to JSON-able object
          representations of NumericsAlertHistory.
        """
        # JSON does not support tuples as keys. Only strings. Therefore, we store
        # the device name, tensor name, and dictionary data within a 3-item list.
        return [
            HistoryTriplet(pair[0], pair[1], history.create_jsonable_history())
            for (pair, history) in self._data.items()
        ]


def extract_numerics_alert(event):
    """Determines whether a health pill event contains bad values.

    A bad value is one of NaN, -Inf, or +Inf.

    Args:
      event: (`Event`) A `tensorflow.Event` proto from `DebugNumericSummary`
        ops.

    Returns:
      An instance of `NumericsAlert`, if bad values are found.
      `None`, if no bad values are found.

    Raises:
      ValueError: if the event does not have the expected tag prefix or the
        debug op name is not the expected debug op name suffix.
    """
    value = event.summary.value[0]
    debugger_plugin_metadata_content = None
    if value.HasField("metadata"):
        plugin_data = value.metadata.plugin_data
        if plugin_data.plugin_name == constants.DEBUGGER_PLUGIN_NAME:
            debugger_plugin_metadata_content = plugin_data.content

    if not debugger_plugin_metadata_content:
        raise ValueError(
            "Event proto input lacks debugger plugin SummaryMetadata."
        )

    debugger_plugin_metadata_content = tf.compat.as_text(
        debugger_plugin_metadata_content
    )
    try:
        content_object = json.loads(debugger_plugin_metadata_content)
        device_name = content_object["device"]
    except (KeyError, ValueError) as e:
        raise ValueError(
            "Could not determine device from JSON string %r, %r"
            % (debugger_plugin_metadata_content, e)
        )

    debug_op_suffix = ":DebugNumericSummary"
    if not value.node_name.endswith(debug_op_suffix):
        raise ValueError(
            "Event proto input does not have the expected debug op suffix %s"
            % debug_op_suffix
        )
    tensor_name = value.node_name[: -len(debug_op_suffix)]

    elements = tf_debug.load_tensor_from_event(event)
    nan_count = elements[constants.NAN_NUMERIC_SUMMARY_OP_INDEX]
    neg_inf_count = elements[constants.NEG_INF_NUMERIC_SUMMARY_OP_INDEX]
    pos_inf_count = elements[constants.POS_INF_NUMERIC_SUMMARY_OP_INDEX]
    if nan_count > 0 or neg_inf_count > 0 or pos_inf_count > 0:
        return NumericsAlert(
            device_name,
            tensor_name,
            event.wall_time,
            nan_count,
            neg_inf_count,
            pos_inf_count,
        )
    return None



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Histogram summaries and TensorFlow operations to create them.

A histogram summary stores a list of buckets. Each bucket is encoded as
a triple `[left_edge, right_edge, count]`. Thus, a full histogram is
encoded as a tensor of dimension `[k, 3]`.

In general, the value of `k` (the number of buckets) will be a constant,
like 30. There are two edge cases: if there is no data, then there are
no buckets (the shape is `[0, 3]`); and if there is data but all points
have the same value, then there is one bucket whose left and right
endpoints are the same (the shape is `[1, 3]`).

NOTE: This module is in beta, and its API is subject to change, but the
data that it stores to disk will be supported forever.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from tensorboard.plugins.histogram import metadata
from tensorboard.plugins.histogram import summary_v2


# Export V2 versions.
histogram = summary_v2.histogram
histogram_pb = summary_v2.histogram_pb


def _buckets(data, bucket_count=None):
    """Create a TensorFlow op to group data into histogram buckets.

    Arguments:
      data: A `Tensor` of any shape. Must be castable to `float64`.
      bucket_count: Optional positive `int` or scalar `int32` `Tensor`.
    Returns:
      A `Tensor` of shape `[k, 3]` and type `float64`. The `i`th row is
      a triple `[left_edge, right_edge, count]` for a single bucket.
      The value of `k` is either `bucket_count` or `1` or `0`.
    """
    # TODO(nickfelt): remove on-demand imports once dep situation is fixed.
    import tensorflow.compat.v1 as tf

    if bucket_count is None:
        bucket_count = summary_v2.DEFAULT_BUCKET_COUNT
    with tf.name_scope(
        "buckets", values=[data, bucket_count]
    ), tf.control_dependencies(
        [tf.assert_scalar(bucket_count), tf.assert_type(bucket_count, tf.int32)]
    ):
        data = tf.reshape(data, shape=[-1])  # flatten
        data = tf.cast(data, tf.float64)
        is_empty = tf.equal(tf.size(input=data), 0)

        def when_empty():
            return tf.constant([], shape=(0, 3), dtype=tf.float64)

        def when_nonempty():
            min_ = tf.reduce_min(input_tensor=data)
            max_ = tf.reduce_max(input_tensor=data)
            range_ = max_ - min_
            is_singular = tf.equal(range_, 0)

            def when_nonsingular():
                bucket_width = range_ / tf.cast(bucket_count, tf.float64)
                offsets = data - min_
                bucket_indices = tf.cast(
                    tf.floor(offsets / bucket_width), dtype=tf.int32
                )
                clamped_indices = tf.minimum(bucket_indices, bucket_count - 1)
                one_hots = tf.one_hot(clamped_indices, depth=bucket_count)
                bucket_counts = tf.cast(
                    tf.reduce_sum(input_tensor=one_hots, axis=0),
                    dtype=tf.float64,
                )
                edges = tf.linspace(min_, max_, bucket_count + 1)
                left_edges = edges[:-1]
                right_edges = edges[1:]
                return tf.transpose(
                    a=tf.stack([left_edges, right_edges, bucket_counts])
                )

            def when_singular():
                center = min_
                bucket_starts = tf.stack([center - 0.5])
                bucket_ends = tf.stack([center + 0.5])
                bucket_counts = tf.stack(
                    [tf.cast(tf.size(input=data), tf.float64)]
                )
                return tf.transpose(
                    a=tf.stack([bucket_starts, bucket_ends, bucket_counts])
                )

            return tf.cond(is_singular, when_singular, when_nonsingular)

        return tf.cond(is_empty, when_empty, when_nonempty)


def op(
    name,
    data,
    bucket_count=None,
    display_name=None,
    description=None,
    collections=None,
):
    """Create a legacy histogram summary op.

    Arguments:
      name: A unique name for the generated summary node.
      data: A `Tensor` of any shape. Must be castable to `float64`.
      bucket_count: Optional positive `int`. The output will have this
        many buckets, except in two edge cases. If there is no data, then
        there are no buckets. If there is data but all points have the
        same value, then there is one bucket whose left and right
        endpoints are the same.
      display_name: Optional name for this summary in TensorBoard, as a
        constant `str`. Defaults to `name`.
      description: Optional long-form description for this summary, as a
        constant `str`. Markdown is supported. Defaults to empty.
      collections: Optional list of graph collections keys. The new
        summary op is added to these collections. Defaults to
        `[Graph Keys.SUMMARIES]`.

    Returns:
      A TensorFlow summary op.
    """
    # TODO(nickfelt): remove on-demand imports once dep situation is fixed.
    import tensorflow.compat.v1 as tf

    if display_name is None:
        display_name = name
    summary_metadata = metadata.create_summary_metadata(
        display_name=display_name, description=description
    )
    with tf.name_scope(name):
        tensor = _buckets(data, bucket_count=bucket_count)
        return tf.summary.tensor_summary(
            name="histogram_summary",
            tensor=tensor,
            collections=collections,
            summary_metadata=summary_metadata,
        )


def pb(name, data, bucket_count=None, display_name=None, description=None):
    """Create a legacy histogram summary protobuf.

    Arguments:
      name: A unique name for the generated summary, including any desired
        name scopes.
      data: A `np.array` or array-like form of any shape. Must have type
        castable to `float`.
      bucket_count: Optional positive `int`. The output will have this
        many buckets, except in two edge cases. If there is no data, then
        there are no buckets. If there is data but all points have the
        same value, then there is one bucket whose left and right
        endpoints are the same.
      display_name: Optional name for this summary in TensorBoard, as a
        `str`. Defaults to `name`.
      description: Optional long-form description for this summary, as a
        `str`. Markdown is supported. Defaults to empty.

    Returns:
      A `tf.Summary` protobuf object.
    """
    # TODO(nickfelt): remove on-demand imports once dep situation is fixed.
    import tensorflow.compat.v1 as tf

    if bucket_count is None:
        bucket_count = summary_v2.DEFAULT_BUCKET_COUNT
    data = np.array(data).flatten().astype(float)
    if data.size == 0:
        buckets = np.array([]).reshape((0, 3))
    else:
        min_ = np.min(data)
        max_ = np.max(data)
        range_ = max_ - min_
        if range_ == 0:
            center = min_
            buckets = np.array([[center - 0.5, center + 0.5, float(data.size)]])
        else:
            bucket_width = range_ / bucket_count
            offsets = data - min_
            bucket_indices = np.floor(offsets / bucket_width).astype(int)
            clamped_indices = np.minimum(bucket_indices, bucket_count - 1)
            one_hots = np.array([clamped_indices]).transpose() == np.arange(
                0, bucket_count
            )  # broadcast
            assert one_hots.shape == (data.size, bucket_count), (
                one_hots.shape,
                (data.size, bucket_count),
            )
            bucket_counts = np.sum(one_hots, axis=0)
            edges = np.linspace(min_, max_, bucket_count + 1)
            left_edges = edges[:-1]
            right_edges = edges[1:]
            buckets = np.array(
                [left_edges, right_edges, bucket_counts]
            ).transpose()
    tensor = tf.make_tensor_proto(buckets, dtype=tf.float64)

    if display_name is None:
        display_name = name
    summary_metadata = metadata.create_summary_metadata(
        display_name=display_name, description=description
    )
    tf_summary_metadata = tf.SummaryMetadata.FromString(
        summary_metadata.SerializeToString()
    )

    summary = tf.Summary()
    summary.value.add(
        tag="%s/histogram_summary" % name,
        metadata=tf_summary_metadata,
        tensor=tensor,
    )
    return summary



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Import router for absl.flags.

See https://github.com/abseil/abseil-py.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging as _logging
import sys as _sys

# go/tf-wildcard-import
from absl.flags import *  # pylint: disable=wildcard-import
import six as _six


# Since we wrap absl.flags DEFINE functions, we need to declare this module
# does not affect key flags.
disclaim_key_flags()  # pylint: disable=undefined-variable


_RENAMED_ARGUMENTS = {
    "flag_name": "name",
    "default_value": "default",
    "docstring": "help",
}


def _wrap_define_function(original_function):
    """Wraps absl.flags's define functions so tf.flags accepts old names."""

    def wrapper(*args, **kwargs):
        """Wrapper function that turns old keyword names to new ones."""
        has_old_names = False
        for old_name, new_name in _six.iteritems(_RENAMED_ARGUMENTS):
            if old_name in kwargs:
                has_old_names = True
                value = kwargs.pop(old_name)
                kwargs[new_name] = value
        if has_old_names:
            _logging.warning(
                "Use of the keyword argument names (flag_name, default_value, "
                "docstring) is deprecated, please use (name, default, help) instead."
            )
        return original_function(*args, **kwargs)

    return wrapper


class _FlagValuesWrapper(object):
    """Wrapper class for absl.flags.FLAGS.

    The difference is that tf.compat.v1.flags.FLAGS implicitly parses
    flags with sys.argv when accessing the FLAGS values before it's
    explicitly parsed, while absl.flags.FLAGS raises an exception.
    """

    def __init__(self, flags_object):
        self.__dict__["__wrapped"] = flags_object

    def __getattribute__(self, name):
        if name == "__dict__":
            return super(_FlagValuesWrapper, self).__getattribute__(name)
        return self.__dict__["__wrapped"].__getattribute__(name)

    def __getattr__(self, name):
        wrapped = self.__dict__["__wrapped"]
        # To maintain backwards compatibility, implicitly parse flags when reading
        # a flag.
        if not wrapped.is_parsed():
            wrapped(_sys.argv)
        return wrapped.__getattr__(name)

    def __setattr__(self, name, value):
        return self.__dict__["__wrapped"].__setattr__(name, value)

    def __delattr__(self, name):
        return self.__dict__["__wrapped"].__delattr__(name)

    def __dir__(self):
        return self.__dict__["__wrapped"].__dir__()

    def __getitem__(self, name):
        return self.__dict__["__wrapped"].__getitem__(name)

    def __setitem__(self, name, flag):
        return self.__dict__["__wrapped"].__setitem__(name, flag)

    def __len__(self):
        return self.__dict__["__wrapped"].__len__()

    def __iter__(self):
        return self.__dict__["__wrapped"].__iter__()

    def __str__(self):
        return self.__dict__["__wrapped"].__str__()

    def __call__(self, *args, **kwargs):
        return self.__dict__["__wrapped"].__call__(*args, **kwargs)


# pylint: disable=invalid-name,used-before-assignment
# absl.flags APIs use `default` as the name of the default value argument.
# Allow the following functions continue to accept `default_value`.
DEFINE_string = _wrap_define_function(DEFINE_string)
DEFINE_boolean = _wrap_define_function(DEFINE_boolean)
DEFINE_bool = DEFINE_boolean
DEFINE_float = _wrap_define_function(DEFINE_float)
DEFINE_integer = _wrap_define_function(DEFINE_integer)
# pylint: enable=invalid-name,used-before-assignment

FLAGS = _FlagValuesWrapper(FLAGS)  # pylint: disable=used-before-assignment



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""IO helper functions."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os
import re

import six

from tensorboard.compat import tf
from tensorboard.util import tb_logging


logger = tb_logging.get_logger()

_ESCAPE_GLOB_CHARACTERS_REGEX = re.compile("([*?[])")


def IsCloudPath(path):
    return (
        path.startswith("gs://")
        or path.startswith("s3://")
        or path.startswith("/cns/")
    )


def PathSeparator(path):
    return "/" if IsCloudPath(path) else os.sep


def IsTensorFlowEventsFile(path):
    """Check the path name to see if it is probably a TF Events file.

    Args:
      path: A file path to check if it is an event file.

    Raises:
      ValueError: If the path is an empty string.

    Returns:
      If path is formatted like a TensorFlowEventsFile. Dummy files such as
        those created with the '.profile-empty' suffixes and meant to hold
        no `Summary` protos are treated as true TensorFlowEventsFiles. For
        background, see: https://github.com/tensorflow/tensorboard/issues/2084.
    """
    if not path:
        raise ValueError("Path must be a nonempty string")
    return "tfevents" in tf.compat.as_str_any(os.path.basename(path))


def IsSummaryEventsFile(path):
    """Check whether the path is probably a TF Events file containing Summary.

    Args:
      path: A file path to check if it is an event file containing `Summary`
        protos.

    Returns:
      If path is formatted like a TensorFlowEventsFile. Dummy files such as
        those created with the '.profile-empty' suffixes and meant to hold
        no `Summary` protos  are treated as `False`. For background, see:
        https://github.com/tensorflow/tensorboard/issues/2084.
    """
    return IsTensorFlowEventsFile(path) and not path.endswith(".profile-empty")


def ListDirectoryAbsolute(directory):
    """Yields all files in the given directory.

    The paths are absolute.
    """
    return (
        os.path.join(directory, path) for path in tf.io.gfile.listdir(directory)
    )


def _EscapeGlobCharacters(path):
    """Escapes the glob characters in a path.

    Python 3 has a glob.escape method, but python 2 lacks it, so we manually
    implement this method.

    Args:
      path: The absolute path to escape.

    Returns:
      The escaped path string.
    """
    drive, path = os.path.splitdrive(path)
    return "%s%s" % (drive, _ESCAPE_GLOB_CHARACTERS_REGEX.sub(r"[\1]", path))


def ListRecursivelyViaGlobbing(top):
    """Recursively lists all files within the directory.

    This method does not list subdirectories (in addition to regular files), and
    the file paths are all absolute. If the directory does not exist, this yields
    nothing.

    This method does so by glob-ing deeper and deeper directories, ie
    foo/*, foo/*/*, foo/*/*/* and so on until all files are listed. All file
    paths are absolute, and this method lists subdirectories too.

    For certain file systems, globbing via this method may prove significantly
    faster than recursively walking a directory. Specifically, TF file systems
    that implement TensorFlow's FileSystem.GetMatchingPaths method could save
    costly disk reads by using this method. However, for other file systems, this
    method might prove slower because the file system performs a walk per call to
    glob (in which case it might as well just perform 1 walk).

    Args:
      top: A path to a directory.

    Yields:
      A (dir_path, file_paths) tuple for each directory/subdirectory.
    """
    current_glob_string = os.path.join(_EscapeGlobCharacters(top), "*")
    level = 0

    while True:
        logger.info("GlobAndListFiles: Starting to glob level %d", level)
        glob = tf.io.gfile.glob(current_glob_string)
        logger.info(
            "GlobAndListFiles: %d files glob-ed at level %d", len(glob), level
        )

        if not glob:
            # This subdirectory level lacks files. Terminate.
            return

        # Map subdirectory to a list of files.
        pairs = collections.defaultdict(list)
        for file_path in glob:
            pairs[os.path.dirname(file_path)].append(file_path)
        for dir_name, file_paths in six.iteritems(pairs):
            yield (dir_name, tuple(file_paths))

        if len(pairs) == 1:
            # If at any point the glob returns files that are all in a single
            # directory, replace the current globbing path with that directory as the
            # literal prefix. This should improve efficiency in cases where a single
            # subdir is significantly deeper than the rest of the sudirs.
            current_glob_string = os.path.join(list(pairs.keys())[0], "*")

        # Iterate to the next level of subdirectories.
        current_glob_string = os.path.join(current_glob_string, "*")
        level += 1


def ListRecursivelyViaWalking(top):
    """Walks a directory tree, yielding (dir_path, file_paths) tuples.

    For each of `top` and its subdirectories, yields a tuple containing the path
    to the directory and the path to each of the contained files.  Note that
    unlike os.Walk()/tf.io.gfile.walk()/ListRecursivelyViaGlobbing, this does not
    list subdirectories. The file paths are all absolute. If the directory does
    not exist, this yields nothing.

    Walking may be incredibly slow on certain file systems.

    Args:
      top: A path to a directory.

    Yields:
      A (dir_path, file_paths) tuple for each directory/subdirectory.
    """
    for dir_path, _, filenames in tf.io.gfile.walk(top, topdown=True):
        yield (
            dir_path,
            (os.path.join(dir_path, filename) for filename in filenames),
        )


def GetLogdirSubdirectories(path):
    """Obtains all subdirectories with events files.

    The order of the subdirectories returned is unspecified. The internal logic
    that determines order varies by scenario.

    Args:
      path: The path to a directory under which to find subdirectories.

    Returns:
      A tuple of absolute paths of all subdirectories each with at least 1 events
      file directly within the subdirectory.

    Raises:
      ValueError: If the path passed to the method exists and is not a directory.
    """
    if not tf.io.gfile.exists(path):
        # No directory to traverse.
        return ()

    if not tf.io.gfile.isdir(path):
        raise ValueError(
            "GetLogdirSubdirectories: path exists and is not a "
            "directory, %s" % path
        )

    if IsCloudPath(path):
        # Glob-ing for files can be significantly faster than recursively
        # walking through directories for some file systems.
        logger.info(
            "GetLogdirSubdirectories: Starting to list directories via glob-ing."
        )
        traversal_method = ListRecursivelyViaGlobbing
    else:
        # For other file systems, the glob-ing based method might be slower because
        # each call to glob could involve performing a recursive walk.
        logger.info(
            "GetLogdirSubdirectories: Starting to list directories via walking."
        )
        traversal_method = ListRecursivelyViaWalking

    return (
        subdir
        for (subdir, files) in traversal_method(path)
        if any(IsTensorFlowEventsFile(f) for f in files)
    )



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Initial server communication to determine session parameters."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from google.protobuf import message
import requests

from absl import logging
from tensorboard import version
from tensorboard.plugins.scalar import metadata as scalars_metadata
from tensorboard.uploader.proto import server_info_pb2


# Request timeout for communicating with remote server.
_REQUEST_TIMEOUT_SECONDS = 10

# Maximum blob size, if not specified by server_info
_DEFAULT_MAX_BLOB_SIZE = 10 * (2 ** 20)  # 10 MiB


def _server_info_request(upload_plugins):
    """Generates a ServerInfoRequest

    Args:
      upload_plugins: List of plugin names requested by the user and to be
        verified by the server.

    Returns:
      A `server_info_pb2.ServerInfoRequest` message.
    """
    request = server_info_pb2.ServerInfoRequest()
    request.version = version.VERSION
    request.plugin_specification.upload_plugins[:] = upload_plugins
    return request


def fetch_server_info(origin, upload_plugins):
    """Fetches server info from a remote server.

    Args:
      origin: The server with which to communicate. Should be a string
        like "https://tensorboard.dev", including protocol, host, and (if
        needed) port.
      upload_plugins: List of plugins names requested by the user and to be
        verified by the server.

    Returns:
      A `server_info_pb2.ServerInfoResponse` message.

    Raises:
      CommunicationError: Upon failure to connect to or successfully
        communicate with the remote server.
    """
    endpoint = "%s/api/uploader" % origin
    server_info_request = _server_info_request(upload_plugins)
    post_body = server_info_request.SerializeToString()
    logging.info("Requested server info: <%r>", server_info_request)
    try:
        response = requests.post(
            endpoint,
            data=post_body,
            timeout=_REQUEST_TIMEOUT_SECONDS,
            headers={"User-Agent": "tensorboard/%s" % version.VERSION},
        )
    except requests.RequestException as e:
        raise CommunicationError("Failed to connect to backend: %s" % e)
    if not response.ok:
        raise CommunicationError(
            "Non-OK status from backend (%d %s): %r"
            % (response.status_code, response.reason, response.content)
        )
    try:
        return server_info_pb2.ServerInfoResponse.FromString(response.content)
    except message.DecodeError as e:
        raise CommunicationError(
            "Corrupt response from backend (%s): %r" % (e, response.content)
        )


def create_server_info(frontend_origin, api_endpoint, upload_plugins):
    """Manually creates server info given a frontend and backend.

    Args:
      frontend_origin: The origin of the TensorBoard.dev frontend, like
        "https://tensorboard.dev" or "http://localhost:8000".
      api_endpoint: As to `server_info_pb2.ApiServer.endpoint`.
      upload_plugins: List of plugin names requested by the user and to be
        verified by the server.

    Returns:
      A `server_info_pb2.ServerInfoResponse` message.
    """
    result = server_info_pb2.ServerInfoResponse()
    result.compatibility.verdict = server_info_pb2.VERDICT_OK
    result.api_server.endpoint = api_endpoint
    url_format = result.url_format
    placeholder = "{{EID}}"
    while placeholder in frontend_origin:
        placeholder = "{%s}" % placeholder
    url_format.template = "%s/experiment/%s/" % (frontend_origin, placeholder)
    url_format.id_placeholder = placeholder
    result.plugin_control.allowed_plugins[:] = upload_plugins
    return result


def experiment_url(server_info, experiment_id):
    """Formats a URL that will resolve to the provided experiment.

    Args:
      server_info: A `server_info_pb2.ServerInfoResponse` message.
      experiment_id: A string; the ID of the experiment to link to.

    Returns:
      A URL resolving to the given experiment, as a string.
    """
    url_format = server_info.url_format
    return url_format.template.replace(url_format.id_placeholder, experiment_id)


def allowed_plugins(server_info):
    """Determines which plugins may upload data.

    This pulls from the `plugin_control` on the `server_info` when that
    submessage is set, else falls back to a default.

    Args:
      server_info: A `server_info_pb2.ServerInfoResponse` message.

    Returns:
      A `frozenset` of plugin names.
    """
    if server_info.HasField("plugin_control"):
        return frozenset(server_info.plugin_control.allowed_plugins)
    else:
        # Old server: gracefully degrade to scalars only, which have
        # been supported since launch. TODO(@wchargin): Promote this
        # branch to an error once we're confident that we won't roll
        # back to old server versions.
        return frozenset((scalars_metadata.PLUGIN_NAME,))


def max_blob_size(server_info):
    """Determines the maximum allowed size for blob uploads.

    This pulls from the `max_blob_size` on the `server_info` when that
    submessage is set, else falls back to a default.

    Note that the RPC endpoint also enforces a limit, which should be kept in
    sync with the one provided by server_info.  Here the purpose is to provide
    a better error message and to avoid attempting the upload if it is expected
    to fail anyway.

    Args:
      server_info: A `server_info_pb2.ServerInfoResponse` message.

    Returns:
      A integer giving the maximum size allowed for blob uploads, in bytes.
    """

    if server_info.HasField("upload_limits"):
        return server_info.upload_limits.max_blob_size
    else:
        # Old server: gracefully degrade to 10 MiB.
        # TODO(@davidsoergel): Promote this branch to an error once we're
        # confident that we won't roll back to old server versions.
        return _DEFAULT_MAX_BLOB_SIZE


class CommunicationError(RuntimeError):
    """Raised upon failure to communicate with the server."""

    pass



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Helper classes for tensor shape inference."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import compat, dtypes
from tensorboard.compat.proto import tensor_shape_pb2


# @tf_export("Dimension")
class Dimension(object):
    """Represents the value of one dimension in a TensorShape."""

    def __init__(self, value):
        """Creates a new Dimension with the given value."""
        if value is None:
            self._value = None
        elif isinstance(value, dtypes.DType):
            raise TypeError("Cannot convert %s to Dimension" % value)
        else:
            self._value = int(value)
            if (
                not isinstance(value, compat.bytes_or_text_types)
                and self._value != value
            ):
                raise ValueError("Ambiguous dimension: %s" % value)
            if self._value < 0:
                raise ValueError("Dimension %d must be >= 0" % self._value)

    def __repr__(self):
        return "Dimension(%s)" % repr(self._value)

    def __str__(self):
        value = self._value
        return "?" if value is None else str(value)

    def __eq__(self, other):
        """Returns true if `other` has the same known value as this
        Dimension."""
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        if self._value is None or other.value is None:
            return None
        return self._value == other.value

    def __ne__(self, other):
        """Returns true if `other` has a different known value from `self`."""
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        if self._value is None or other.value is None:
            return None
        return self._value != other.value

    def __int__(self):
        return self._value

    # This is needed for Windows.
    # See https://github.com/tensorflow/tensorflow/pull/9780
    def __long__(self):
        return self._value

    def __index__(self):
        # Allow use in Python 3 range
        return self._value

    @property
    def value(self):
        """The value of this dimension, or None if it is unknown."""
        return self._value

    def is_convertible_with(self, other):
        """Returns true if `other` is convertible with this Dimension.

        Two known Dimensions are convertible if they have the same value.
        An unknown Dimension is convertible with all other Dimensions.

        Args:
          other: Another Dimension.

        Returns:
          True if this Dimension and `other` are convertible.
        """
        other = as_dimension(other)
        return (
            self._value is None
            or other.value is None
            or self._value == other.value
        )

    def assert_is_convertible_with(self, other):
        """Raises an exception if `other` is not convertible with this
        Dimension.

        Args:
          other: Another Dimension.

        Raises:
          ValueError: If `self` and `other` are not convertible (see
            is_convertible_with).
        """
        if not self.is_convertible_with(other):
            raise ValueError(
                "Dimensions %s and %s are not convertible" % (self, other)
            )

    def merge_with(self, other):
        """Returns a Dimension that combines the information in `self` and
        `other`.

        Dimensions are combined as follows:

        ```python
        tf.Dimension(n)   .merge_with(tf.Dimension(n))    == tf.Dimension(n)
        tf.Dimension(n)   .merge_with(tf.Dimension(None)) == tf.Dimension(n)
        tf.Dimension(None).merge_with(tf.Dimension(n))    == tf.Dimension(n)
        tf.Dimension(None).merge_with(tf.Dimension(None)) == tf.Dimension(None)
        tf.Dimension(n)   .merge_with(tf.Dimension(m))  # raises ValueError for n != m
        ```

        Args:
          other: Another Dimension.

        Returns:
          A Dimension containing the combined information of `self` and
          `other`.

        Raises:
          ValueError: If `self` and `other` are not convertible (see
            is_convertible_with).
        """
        other = as_dimension(other)
        self.assert_is_convertible_with(other)
        if self._value is None:
            return Dimension(other.value)
        else:
            return Dimension(self._value)

    def __add__(self, other):
        """Returns the sum of `self` and `other`.

        Dimensions are summed as follows:

        ```python
        tf.Dimension(m)    + tf.Dimension(n)    == tf.Dimension(m + n)
        tf.Dimension(m)    + tf.Dimension(None) == tf.Dimension(None)
        tf.Dimension(None) + tf.Dimension(n)    == tf.Dimension(None)
        tf.Dimension(None) + tf.Dimension(None) == tf.Dimension(None)
        ```

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A Dimension whose value is the sum of `self` and `other`.
        """
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value + other.value)

    def __radd__(self, other):
        """Returns the sum of `other` and `self`.

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A Dimension whose value is the sum of `self` and `other`.
        """
        return self + other

    def __sub__(self, other):
        """Returns the subtraction of `other` from `self`.

        Dimensions are subtracted as follows:

        ```python
        tf.Dimension(m)    - tf.Dimension(n)    == tf.Dimension(m - n)
        tf.Dimension(m)    - tf.Dimension(None) == tf.Dimension(None)
        tf.Dimension(None) - tf.Dimension(n)    == tf.Dimension(None)
        tf.Dimension(None) - tf.Dimension(None) == tf.Dimension(None)
        ```

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A Dimension whose value is the subtraction of `other` from `self`.
        """
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value - other.value)

    def __rsub__(self, other):
        """Returns the subtraction of `self` from `other`.

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A Dimension whose value is the subtraction of `self` from `other`.
        """
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(other.value - self._value)

    def __mul__(self, other):
        """Returns the product of `self` and `other`.

        Dimensions are summed as follows:

        ```python
        tf.Dimension(m)    * tf.Dimension(n)    == tf.Dimension(m * n)
        tf.Dimension(m)    * tf.Dimension(None) == tf.Dimension(None)
        tf.Dimension(None) * tf.Dimension(n)    == tf.Dimension(None)
        tf.Dimension(None) * tf.Dimension(None) == tf.Dimension(None)
        ```

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A Dimension whose value is the product of `self` and `other`.
        """
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented

        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value * other.value)

    def __rmul__(self, other):
        """Returns the product of `self` and `other`.

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A Dimension whose value is the product of `self` and `other`.
        """
        return self * other

    def __floordiv__(self, other):
        """Returns the quotient of `self` and `other` rounded down.

        Dimensions are divided as follows:

        ```python
        tf.Dimension(m)    // tf.Dimension(n)    == tf.Dimension(m // n)
        tf.Dimension(m)    // tf.Dimension(None) == tf.Dimension(None)
        tf.Dimension(None) // tf.Dimension(n)    == tf.Dimension(None)
        tf.Dimension(None) // tf.Dimension(None) == tf.Dimension(None)
        ```

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A `Dimension` whose value is the integer quotient of `self` and `other`.
        """
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value // other.value)

    def __rfloordiv__(self, other):
        """Returns the quotient of `other` and `self` rounded down.

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A `Dimension` whose value is the integer quotient of `self` and `other`.
        """
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(other.value // self._value)

    def __div__(self, other):
        """DEPRECATED: Use `__floordiv__` via `x // y` instead.

        This function exists only for backwards convertibility purposes; new code
        should use `__floordiv__` via the syntax `x // y`.  Using `x // y`
        communicates clearly that the result rounds down, and is forward convertible
        to Python 3.

        Args:
          other: Another `Dimension`.

        Returns:
          A `Dimension` whose value is the integer quotient of `self` and `other`.
        """
        return self // other

    def __mod__(self, other):
        """Returns `self` modulo `other`.

        Dimension moduli are computed as follows:

        ```python
        tf.Dimension(m)    % tf.Dimension(n)    == tf.Dimension(m % n)
        tf.Dimension(m)    % tf.Dimension(None) == tf.Dimension(None)
        tf.Dimension(None) % tf.Dimension(n)    == tf.Dimension(None)
        tf.Dimension(None) % tf.Dimension(None) == tf.Dimension(None)
        ```

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A Dimension whose value is `self` modulo `other`.
        """
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        if self._value is None or other.value is None:
            return Dimension(None)
        else:
            return Dimension(self._value % other.value)

    def __rmod__(self, other):
        """Returns `other` modulo `self`.

        Args:
          other: Another Dimension, or a value accepted by `as_dimension`.

        Returns:
          A Dimension whose value is `other` modulo `self`.
        """
        try:
            other = as_dimension(other)
        except (TypeError, ValueError):
            return NotImplemented
        return other % self

    def __lt__(self, other):
        """Returns True if `self` is known to be less than `other`.

        Dimensions are compared as follows:

        ```python
        (tf.Dimension(m)    < tf.Dimension(n))    == (m < n)
        (tf.Dimension(m)    < tf.Dimension(None)) == None
        (tf.Dimension(None) < tf.Dimension(n))    == None
        (tf.Dimension(None) < tf.Dimension(None)) == None
        ```

        Args:
          other: Another Dimension.

        Returns:
          The value of `self.value < other.value` if both are known, otherwise
          None.
        """
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return None
        else:
            return self._value < other.value

    def __le__(self, other):
        """Returns True if `self` is known to be less than or equal to `other`.

        Dimensions are compared as follows:

        ```python
        (tf.Dimension(m)    <= tf.Dimension(n))    == (m <= n)
        (tf.Dimension(m)    <= tf.Dimension(None)) == None
        (tf.Dimension(None) <= tf.Dimension(n))    == None
        (tf.Dimension(None) <= tf.Dimension(None)) == None
        ```

        Args:
          other: Another Dimension.

        Returns:
          The value of `self.value <= other.value` if both are known, otherwise
          None.
        """
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return None
        else:
            return self._value <= other.value

    def __gt__(self, other):
        """Returns True if `self` is known to be greater than `other`.

        Dimensions are compared as follows:

        ```python
        (tf.Dimension(m)    > tf.Dimension(n))    == (m > n)
        (tf.Dimension(m)    > tf.Dimension(None)) == None
        (tf.Dimension(None) > tf.Dimension(n))    == None
        (tf.Dimension(None) > tf.Dimension(None)) == None
        ```

        Args:
          other: Another Dimension.

        Returns:
          The value of `self.value > other.value` if both are known, otherwise
          None.
        """
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return None
        else:
            return self._value > other.value

    def __ge__(self, other):
        """Returns True if `self` is known to be greater than or equal to
        `other`.

        Dimensions are compared as follows:

        ```python
        (tf.Dimension(m)    >= tf.Dimension(n))    == (m >= n)
        (tf.Dimension(m)    >= tf.Dimension(None)) == None
        (tf.Dimension(None) >= tf.Dimension(n))    == None
        (tf.Dimension(None) >= tf.Dimension(None)) == None
        ```

        Args:
          other: Another Dimension.

        Returns:
          The value of `self.value >= other.value` if both are known, otherwise
          None.
        """
        other = as_dimension(other)
        if self._value is None or other.value is None:
            return None
        else:
            return self._value >= other.value

    def __reduce__(self):
        return Dimension, (self._value,)


def as_dimension(value):
    """Converts the given value to a Dimension.

    A Dimension input will be returned unmodified.
    An input of `None` will be converted to an unknown Dimension.
    An integer input will be converted to a Dimension with that value.

    Args:
      value: The value to be converted.

    Returns:
      A Dimension corresponding to the given value.
    """
    if isinstance(value, Dimension):
        return value
    else:
        return Dimension(value)


# @tf_export("TensorShape")
class TensorShape(object):
    """Represents the shape of a `Tensor`.

    A `TensorShape` represents a possibly-partial shape specification for a
    `Tensor`. It may be one of the following:

    * *Fully-known shape:* has a known number of dimensions and a known size
      for each dimension. e.g. `TensorShape([16, 256])`
    * *Partially-known shape:* has a known number of dimensions, and an unknown
      size for one or more dimension. e.g. `TensorShape([None, 256])`
    * *Unknown shape:* has an unknown number of dimensions, and an unknown
      size in all dimensions. e.g. `TensorShape(None)`

    If a tensor is produced by an operation of type `"Foo"`, its shape
    may be inferred if there is a registered shape function for
    `"Foo"`. See @{$adding_an_op#shape-functions-in-c$`Shape functions in C++`}
    for details of shape functions and how to register them. Alternatively,
    the shape may be set explicitly using @{tf.Tensor.set_shape}.
    """

    def __init__(self, dims):
        """Creates a new TensorShape with the given dimensions.

        Args:
          dims: A list of Dimensions, or None if the shape is unspecified.
            DEPRECATED: A single integer is treated as a singleton list.

        Raises:
          TypeError: If dims cannot be converted to a list of dimensions.
        """
        # TODO(irving): Eliminate the single integer special case.
        if dims is None:
            self._dims = None
        elif isinstance(dims, compat.bytes_or_text_types):
            raise TypeError(
                "A string has ambiguous TensorShape, please wrap in a "
                "list or convert to an int: %s" % dims
            )
        elif isinstance(dims, tensor_shape_pb2.TensorShapeProto):
            if dims.unknown_rank:
                self._dims = None
            else:
                self._dims = [
                    # Protos store variable-size dimensions as -1
                    as_dimension(dim.size if dim.size != -1 else None)
                    for dim in dims.dim
                ]
        elif isinstance(dims, TensorShape):
            self._dims = dims.dims
        else:
            try:
                dims_iter = iter(dims)
            except TypeError:
                # Treat as a singleton dimension
                self._dims = [as_dimension(dims)]
            else:
                # Got a list of dimensions
                self._dims = [as_dimension(d) for d in dims_iter]
        self._ndims = None

    def __repr__(self):
        return "TensorShape(%r)" % self._dims

    def __str__(self):
        if self.ndims is None:
            return "<unknown>"
        elif self.ndims == 1:
            return "(%s,)" % self._dims[0]
        else:
            return "(%s)" % ", ".join(str(d) for d in self._dims)

    @property
    def dims(self):
        """Returns a list of Dimensions, or None if the shape is
        unspecified."""
        return self._dims

    @dims.setter
    def dims(self, dims):
        self._dims = dims
        self._ndims = None

    @property
    def ndims(self):
        """Returns the rank of this shape, or None if it is unspecified."""
        if self._dims is None:
            return None
        else:
            if self._ndims is None:
                self._ndims = len(self._dims)
            return self._ndims

    def __len__(self):
        """Returns the rank of this shape, or raises ValueError if
        unspecified."""
        if self._dims is None:
            raise ValueError(
                "Cannot take the length of Shape with unknown rank."
            )
        return self.ndims

    def __bool__(self):
        """Returns True if this shape contains non-zero information."""
        return self._dims is not None

    # Python 3 wants __bool__, Python 2.7 wants __nonzero__
    __nonzero__ = __bool__

    def __iter__(self):
        """Returns `self.dims` if the rank is known, otherwise raises
        ValueError."""
        if self._dims is None:
            raise ValueError("Cannot iterate over a shape with unknown rank.")
        else:
            return iter(self._dims)

    def __getitem__(self, key):
        """Returns the value of a dimension or a shape, depending on the key.

        Args:
          key: If `key` is an integer, returns the dimension at that index;
            otherwise if `key` is a slice, returns a TensorShape whose
            dimensions are those selected by the slice from `self`.

        Returns:
          A dimension if `key` is an integer, or a `TensorShape` if `key` is a
          slice.

        Raises:
          ValueError: If `key` is a slice, and any of its elements are negative, or
            if `self` is completely unknown and the step is set.
        """
        if self._dims is not None:
            if isinstance(key, slice):
                return TensorShape(self._dims[key])
            else:
                return self._dims[key]
        else:
            if isinstance(key, slice):
                start = key.start if key.start is not None else 0
                stop = key.stop

                if key.step is not None:
                    # TODO(mrry): Handle these maybe.
                    raise ValueError("Steps are not yet handled")
                if stop is None:
                    # NOTE(mrry): This implies that TensorShape(None) is convertible with
                    # TensorShape(None)[1:], which is obviously not true. It would be
                    # possible to track the number of dimensions symbolically,
                    # and perhaps we should do that.
                    return unknown_shape()
                elif start < 0 or stop < 0:
                    # TODO(mrry): Handle this better, as it will be useful for handling
                    # suffixes of otherwise unknown shapes.
                    return unknown_shape()
                else:
                    return unknown_shape(ndims=stop - start)
            else:
                return Dimension(None)

    def num_elements(self):
        """Returns the total number of elements, or none for incomplete
        shapes."""
        if self.is_fully_defined():
            size = 1
            for dim in self._dims:
                size *= dim.value
            return size
        else:
            return None

    def merge_with(self, other):
        """Returns a `TensorShape` combining the information in `self` and
        `other`.

        The dimensions in `self` and `other` are merged elementwise,
        according to the rules defined for `Dimension.merge_with()`.

        Args:
          other: Another `TensorShape`.

        Returns:
          A `TensorShape` containing the combined information of `self` and
          `other`.

        Raises:
          ValueError: If `self` and `other` are not convertible.
        """
        other = as_shape(other)
        if self._dims is None:
            return other
        else:
            try:
                self.assert_same_rank(other)
                new_dims = []
                for i, dim in enumerate(self._dims):
                    new_dims.append(dim.merge_with(other[i]))
                return TensorShape(new_dims)
            except ValueError:
                raise ValueError(
                    "Shapes %s and %s are not convertible" % (self, other)
                )

    def concatenate(self, other):
        """Returns the concatenation of the dimension in `self` and `other`.

        *N.B.* If either `self` or `other` is completely unknown,
        concatenation will discard information about the other shape. In
        future, we might support concatenation that preserves this
        information for use with slicing.

        Args:
          other: Another `TensorShape`.

        Returns:
          A `TensorShape` whose dimensions are the concatenation of the
          dimensions in `self` and `other`.
        """
        # TODO(mrry): Handle the case where we concatenate a known shape with a
        # completely unknown shape, so that we can use the partial information.
        other = as_shape(other)
        if self._dims is None or other.dims is None:
            return unknown_shape()
        else:
            return TensorShape(self._dims + other.dims)

    def assert_same_rank(self, other):
        """Raises an exception if `self` and `other` do not have convertible
        ranks.

        Args:
          other: Another `TensorShape`.

        Raises:
          ValueError: If `self` and `other` do not represent shapes with the
            same rank.
        """
        other = as_shape(other)
        if self.ndims is not None and other.ndims is not None:
            if self.ndims != other.ndims:
                raise ValueError(
                    "Shapes %s and %s must have the same rank" % (self, other)
                )

    def assert_has_rank(self, rank):
        """Raises an exception if `self` is not convertible with the given
        `rank`.

        Args:
          rank: An integer.

        Raises:
          ValueError: If `self` does not represent a shape with the given `rank`.
        """
        if self.ndims not in (None, rank):
            raise ValueError("Shape %s must have rank %d" % (self, rank))

    def with_rank(self, rank):
        """Returns a shape based on `self` with the given rank.

        This method promotes a completely unknown shape to one with a
        known rank.

        Args:
          rank: An integer.

        Returns:
          A shape that is at least as specific as `self` with the given rank.

        Raises:
          ValueError: If `self` does not represent a shape with the given `rank`.
        """
        try:
            return self.merge_with(unknown_shape(ndims=rank))
        except ValueError:
            raise ValueError("Shape %s must have rank %d" % (self, rank))

    def with_rank_at_least(self, rank):
        """Returns a shape based on `self` with at least the given rank.

        Args:
          rank: An integer.

        Returns:
          A shape that is at least as specific as `self` with at least the given
          rank.

        Raises:
          ValueError: If `self` does not represent a shape with at least the given
            `rank`.
        """
        if self.ndims is not None and self.ndims < rank:
            raise ValueError(
                "Shape %s must have rank at least %d" % (self, rank)
            )
        else:
            return self

    def with_rank_at_most(self, rank):
        """Returns a shape based on `self` with at most the given rank.

        Args:
          rank: An integer.

        Returns:
          A shape that is at least as specific as `self` with at most the given
          rank.

        Raises:
          ValueError: If `self` does not represent a shape with at most the given
            `rank`.
        """
        if self.ndims is not None and self.ndims > rank:
            raise ValueError(
                "Shape %s must have rank at most %d" % (self, rank)
            )
        else:
            return self

    def is_convertible_with(self, other):
        """Returns True iff `self` is convertible with `other`.

        Two possibly-partially-defined shapes are convertible if there
        exists a fully-defined shape that both shapes can represent. Thus,
        convertibility allows the shape inference code to reason about
        partially-defined shapes. For example:

        * TensorShape(None) is convertible with all shapes.

        * TensorShape([None, None]) is convertible with all two-dimensional
          shapes, such as TensorShape([32, 784]), and also TensorShape(None). It is
          not convertible with, for example, TensorShape([None]) or
          TensorShape([None, None, None]).

        * TensorShape([32, None]) is convertible with all two-dimensional shapes
          with size 32 in the 0th dimension, and also TensorShape([None, None])
          and TensorShape(None). It is not convertible with, for example,
          TensorShape([32]), TensorShape([32, None, 1]) or TensorShape([64, None]).

        * TensorShape([32, 784]) is convertible with itself, and also
          TensorShape([32, None]), TensorShape([None, 784]), TensorShape([None,
          None]) and TensorShape(None). It is not convertible with, for example,
          TensorShape([32, 1, 784]) or TensorShape([None]).

        The convertibility relation is reflexive and symmetric, but not
        transitive. For example, TensorShape([32, 784]) is convertible with
        TensorShape(None), and TensorShape(None) is convertible with
        TensorShape([4, 4]), but TensorShape([32, 784]) is not convertible with
        TensorShape([4, 4]).

        Args:
          other: Another TensorShape.

        Returns:
          True iff `self` is convertible with `other`.
        """
        other = as_shape(other)
        if self._dims is not None and other.dims is not None:
            if self.ndims != other.ndims:
                return False
            for x_dim, y_dim in zip(self._dims, other.dims):
                if not x_dim.is_convertible_with(y_dim):
                    return False
        return True

    def assert_is_convertible_with(self, other):
        """Raises exception if `self` and `other` do not represent the same
        shape.

        This method can be used to assert that there exists a shape that both
        `self` and `other` represent.

        Args:
          other: Another TensorShape.

        Raises:
          ValueError: If `self` and `other` do not represent the same shape.
        """
        if not self.is_convertible_with(other):
            raise ValueError(
                "Shapes %s and %s are inconvertible" % (self, other)
            )

    def most_specific_convertible_shape(self, other):
        """Returns the most specific TensorShape convertible with `self` and
        `other`.

        * TensorShape([None, 1]) is the most specific TensorShape convertible with
          both TensorShape([2, 1]) and TensorShape([5, 1]). Note that
          TensorShape(None) is also convertible with above mentioned TensorShapes.

        * TensorShape([1, 2, 3]) is the most specific TensorShape convertible with
          both TensorShape([1, 2, 3]) and TensorShape([1, 2, 3]). There are more
          less specific TensorShapes convertible with above mentioned TensorShapes,
          e.g. TensorShape([1, 2, None]), TensorShape(None).

        Args:
          other: Another `TensorShape`.

        Returns:
          A `TensorShape` which is the most specific convertible shape of `self`
          and `other`.
        """

        other = as_shape(other)
        if (
            self._dims is None
            or other.dims is None
            or self.ndims != other.ndims
        ):
            return unknown_shape()

        dims = [(Dimension(None))] * self.ndims
        for i, (d1, d2) in enumerate(zip(self._dims, other.dims)):
            if d1 is not None and d2 is not None and d1 == d2:
                dims[i] = d1
        return TensorShape(dims)

    def is_fully_defined(self):
        """Returns True iff `self` is fully defined in every dimension."""
        return self._dims is not None and all(
            dim.value is not None for dim in self._dims
        )

    def assert_is_fully_defined(self):
        """Raises an exception if `self` is not fully defined in every
        dimension.

        Raises:
          ValueError: If `self` does not have a known value for every dimension.
        """
        if not self.is_fully_defined():
            raise ValueError("Shape %s is not fully defined" % self)

    def as_list(self):
        """Returns a list of integers or `None` for each dimension.

        Returns:
          A list of integers or `None` for each dimension.

        Raises:
          ValueError: If `self` is an unknown shape with an unknown rank.
        """
        if self._dims is None:
            raise ValueError(
                "as_list() is not defined on an unknown TensorShape."
            )
        return [dim.value for dim in self._dims]

    def as_proto(self):
        """Returns this shape as a `TensorShapeProto`."""
        if self._dims is None:
            return tensor_shape_pb2.TensorShapeProto(unknown_rank=True)
        else:
            return tensor_shape_pb2.TensorShapeProto(
                dim=[
                    tensor_shape_pb2.TensorShapeProto.Dim(
                        size=-1 if d.value is None else d.value
                    )
                    for d in self._dims
                ]
            )

    def __eq__(self, other):
        """Returns True if `self` is equivalent to `other`."""
        try:
            other = as_shape(other)
        except TypeError:
            return NotImplemented
        return self._dims == other.dims

    def __ne__(self, other):
        """Returns True if `self` is known to be different from `other`."""
        try:
            other = as_shape(other)
        except TypeError:
            return NotImplemented
        if self.ndims is None or other.ndims is None:
            raise ValueError(
                "The inequality of unknown TensorShapes is undefined."
            )
        if self.ndims != other.ndims:
            return True
        return self._dims != other.dims

    def __reduce__(self):
        return TensorShape, (self._dims,)


def as_shape(shape):
    """Converts the given object to a TensorShape."""
    if isinstance(shape, TensorShape):
        return shape
    else:
        return TensorShape(shape)


def unknown_shape(ndims=None):
    """Returns an unknown TensorShape, optionally with a known rank.

    Args:
      ndims: (Optional) If specified, the number of dimensions in the shape.

    Returns:
      An unknown TensorShape.
    """
    if ndims is None:
        return TensorShape(None)
    else:
        return TensorShape([Dimension(None)] * ndims)


_SCALAR_SHAPE = TensorShape([])


def scalar():
    """Returns a shape representing a scalar."""
    return _SCALAR_SHAPE


def vector(length):
    """Returns a shape representing a vector.

    Args:
      length: The length of the vector, which may be None if unknown.

    Returns:
      A TensorShape representing a vector of the given length.
    """
    return TensorShape([length])


def matrix(rows, cols):
    """Returns a shape representing a matrix.

    Args:
      rows: The number of rows in the matrix, which may be None if unknown.
      cols: The number of columns in the matrix, which may be None if unknown.

    Returns:
      A TensorShape representing a matrix of the given size.
    """
    return TensorShape([rows, cols])



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The TensorBoard Audio plugin."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from six.moves import urllib
from werkzeug import wrappers

from tensorboard import plugin_util
from tensorboard.backend import http_util
from tensorboard.compat import tf
from tensorboard.plugins import base_plugin
from tensorboard.plugins.audio import metadata
from tensorboard.util import tensor_util


_DEFAULT_MIME_TYPE = "application/octet-stream"
_MIME_TYPES = {
    metadata.Encoding.Value("WAV"): "audio/wav",
}


class AudioPlugin(base_plugin.TBPlugin):
    """Audio Plugin for TensorBoard."""

    plugin_name = metadata.PLUGIN_NAME

    def __init__(self, context):
        """Instantiates AudioPlugin via TensorBoard core.

        Args:
          context: A base_plugin.TBContext instance.
        """
        self._multiplexer = context.multiplexer

    def get_plugin_apps(self):
        return {
            "/audio": self._serve_audio_metadata,
            "/individualAudio": self._serve_individual_audio,
            "/tags": self._serve_tags,
        }

    def is_active(self):
        """The audio plugin is active iff any run has at least one relevant
        tag."""
        if not self._multiplexer:
            return False
        return bool(
            self._multiplexer.PluginRunToTagToContent(metadata.PLUGIN_NAME)
        )

    def frontend_metadata(self):
        return base_plugin.FrontendMetadata(element_name="tf-audio-dashboard")

    def _index_impl(self):
        """Return information about the tags in each run.

        Result is a dictionary of the form

            {
              "runName1": {
                "tagName1": {
                  "displayName": "The first tag",
                  "description": "<p>Long ago there was just one tag...</p>",
                  "samples": 3
                },
                "tagName2": ...,
                ...
              },
              "runName2": ...,
              ...
            }

        For each tag, `samples` is the greatest number of audio clips that
        appear at any particular step. (It's not related to "samples of a
        waveform.") For example, if for tag `minibatch_input` there are
        five audio clips at step 0 and ten audio clips at step 1, then the
        dictionary for `"minibatch_input"` will contain `"samples": 10`.
        """
        runs = self._multiplexer.Runs()
        result = {run: {} for run in runs}

        mapping = self._multiplexer.PluginRunToTagToContent(
            metadata.PLUGIN_NAME
        )
        for (run, tag_to_content) in six.iteritems(mapping):
            for tag in tag_to_content:
                summary_metadata = self._multiplexer.SummaryMetadata(run, tag)
                tensor_events = self._multiplexer.Tensors(run, tag)
                samples = max(
                    [
                        self._number_of_samples(event.tensor_proto)
                        for event in tensor_events
                    ]
                    + [0]
                )
                result[run][tag] = {
                    "displayName": summary_metadata.display_name,
                    "description": plugin_util.markdown_to_safe_html(
                        summary_metadata.summary_description
                    ),
                    "samples": samples,
                }

        return result

    def _number_of_samples(self, tensor_proto):
        """Count the number of samples of an audio TensorProto."""
        # We directly inspect the `tensor_shape` of the proto instead of
        # using the preferred `tensor_util.make_ndarray(...).shape`, because
        # these protos can contain a large amount of encoded audio data,
        # and we don't want to have to convert them all to numpy arrays
        # just to look at their shape.
        return tensor_proto.tensor_shape.dim[0].size

    def _filter_by_sample(self, tensor_events, sample):
        return [
            tensor_event
            for tensor_event in tensor_events
            if self._number_of_samples(tensor_event.tensor_proto) > sample
        ]

    @wrappers.Request.application
    def _serve_audio_metadata(self, request):
        """Given a tag and list of runs, serve a list of metadata for audio.

        Note that the actual audio data are not sent; instead, we respond
        with URLs to the audio. The frontend should treat these URLs as
        opaque and should not try to parse information about them or
        generate them itself, as the format may change.

        Args:
          request: A werkzeug.wrappers.Request object.

        Returns:
          A werkzeug.Response application.
        """
        tag = request.args.get("tag")
        run = request.args.get("run")
        sample = int(request.args.get("sample", 0))

        events = self._multiplexer.Tensors(run, tag)
        try:
            response = self._audio_response_for_run(events, run, tag, sample)
        except KeyError:
            return http_util.Respond(
                request, "Invalid run or tag", "text/plain", code=400
            )
        return http_util.Respond(request, response, "application/json")

    def _audio_response_for_run(self, tensor_events, run, tag, sample):
        """Builds a JSON-serializable object with information about audio.

        Args:
          tensor_events: A list of image event_accumulator.TensorEvent objects.
          run: The name of the run.
          tag: The name of the tag the audio entries all belong to.
          sample: The zero-indexed sample of the audio sample for which to
          retrieve information. For instance, setting `sample` to `2` will
            fetch information about only the third audio clip of each batch,
            and steps with fewer than three audio clips will be omitted from
            the results.

        Returns:
          A list of dictionaries containing the wall time, step, URL, width, and
          height for each audio entry.
        """
        response = []
        index = 0
        filtered_events = self._filter_by_sample(tensor_events, sample)
        content_type = self._get_mime_type(run, tag)
        for (index, tensor_event) in enumerate(filtered_events):
            data = tensor_util.make_ndarray(tensor_event.tensor_proto)
            label = data[sample, 1]
            response.append(
                {
                    "wall_time": tensor_event.wall_time,
                    "step": tensor_event.step,
                    "label": plugin_util.markdown_to_safe_html(label),
                    "contentType": content_type,
                    "query": self._query_for_individual_audio(
                        run, tag, sample, index
                    ),
                }
            )
        return response

    def _query_for_individual_audio(self, run, tag, sample, index):
        """Builds a URL for accessing the specified audio.

        This should be kept in sync with _serve_audio_metadata. Note that the URL is
        *not* guaranteed to always return the same audio, since audio may be
        unloaded from the reservoir as new audio entries come in.

        Args:
          run: The name of the run.
          tag: The tag.
          index: The index of the audio entry. Negative values are OK.

        Returns:
          A string representation of a URL that will load the index-th sampled audio
          in the given run with the given tag.
        """
        query_string = urllib.parse.urlencode(
            {"run": run, "tag": tag, "sample": sample, "index": index,}
        )
        return query_string

    def _get_mime_type(self, run, tag):
        content = self._multiplexer.SummaryMetadata(
            run, tag
        ).plugin_data.content
        parsed = metadata.parse_plugin_metadata(content)
        return _MIME_TYPES.get(parsed.encoding, _DEFAULT_MIME_TYPE)

    @wrappers.Request.application
    def _serve_individual_audio(self, request):
        """Serve encoded audio data."""
        tag = request.args.get("tag")
        run = request.args.get("run")
        index = int(request.args.get("index", "0"))
        sample = int(request.args.get("sample", "0"))
        try:
            events = self._filter_by_sample(
                self._multiplexer.Tensors(run, tag), sample
            )
            data = tensor_util.make_ndarray(events[index].tensor_proto)[
                sample, 0
            ]
        except (KeyError, IndexError):
            return http_util.Respond(
                request,
                "Invalid run, tag, index, or sample",
                "text/plain",
                code=400,
            )
        mime_type = self._get_mime_type(run, tag)
        return http_util.Respond(request, data, mime_type)

    @wrappers.Request.application
    def _serve_tags(self, request):
        index = self._index_impl()
        return http_util.Respond(request, index, "application/json")



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Contains error codes defined in tensorflow/core/lib/core/error_codes.proto

# Not an error; returned on success
OK = 0

# The operation was cancelled (typically by the caller).
CANCELLED = 1

"""
Unknown error.  An example of where this error may be returned is
if a Status value received from another address space belongs to
an error-space that is not known in this address space.  Also
errors raised by APIs that do not return enough error information
may be converted to this error.
"""
UNKNOWN = 2

"""
Client specified an invalid argument.  Note that this differs
from FAILED_PRECONDITION.  INVALID_ARGUMENT indicates arguments
that are problematic regardless of the state of the system
(e.g., a malformed file name).
"""
INVALID_ARGUMENT = 3

"""
Deadline expired before operation could complete.  For operations
that change the state of the system, this error may be returned
even if the operation has completed successfully.  For example, a
successful response from a server could have been delayed long
enough for the deadline to expire.
"""
DEADLINE_EXCEEDED = 4

"""
Some requested entity (e.g., file or directory) was not found.
For privacy reasons, this code *may* be returned when the client
does not have the access right to the entity.
"""
NOT_FOUND = 5

"""
Some entity that we attempted to create (e.g., file or directory)
already exists.
"""
ALREADY_EXISTS = 6

"""
The caller does not have permission to execute the specified
operation.  PERMISSION_DENIED must not be used for rejections
caused by exhausting some resource (use RESOURCE_EXHAUSTED
instead for those errors).  PERMISSION_DENIED must not be
used if the caller can not be identified (use UNAUTHENTICATED
instead for those errors).
"""
PERMISSION_DENIED = 7

"""
Some resource has been exhausted, perhaps a per-user quota, or
perhaps the entire file system is out of space.
"""
RESOURCE_EXHAUSTED = 8

"""
Operation was rejected because the system is not in a state
required for the operation's execution.  For example, directory
to be deleted may be non-empty, an rmdir operation is applied to
a non-directory, etc.
A litmus test that may help a service implementor in deciding
between FAILED_PRECONDITION, ABORTED, and UNAVAILABLE:
(a) Use UNAVAILABLE if the client can retry just the failing call.
(b) Use ABORTED if the client should retry at a higher-level
    (e.g., restarting a read-modify-write sequence).
(c) Use FAILED_PRECONDITION if the client should not retry until
    the system state has been explicitly fixed.  E.g., if an "rmdir"
    fails because the directory is non-empty, FAILED_PRECONDITION
    should be returned since the client should not retry unless
    they have first fixed up the directory by deleting files from it.
(d) Use FAILED_PRECONDITION if the client performs conditional
    REST Get/Update/Delete on a resource and the resource on the
    server does not match the condition. E.g., conflicting
    read-modify-write on the same resource.
"""
FAILED_PRECONDITION = 9

"""
The operation was aborted, typically due to a concurrency issue
like sequencer check failures, transaction aborts, etc.

See litmus test above for deciding between FAILED_PRECONDITION,
ABORTED, and UNAVAILABLE.
"""
ABORTED = 10

"""
Operation tried to iterate past the valid input range.  E.g., seeking or
reading past end of file.

Unlike INVALID_ARGUMENT, this error indicates a problem that may
be fixed if the system state changes. For example, a 32-bit file
system will generate INVALID_ARGUMENT if asked to read at an
offset that is not in the range [0,2^32-1], but it will generate
OUT_OF_RANGE if asked to read from an offset past the current
file size.

There is a fair bit of overlap between FAILED_PRECONDITION and
OUT_OF_RANGE.  We recommend using OUT_OF_RANGE (the more specific
error) when it applies so that callers who are iterating through
a space can easily look for an OUT_OF_RANGE error to detect when
they are done.
"""
OUT_OF_RANGE = 11

# Operation is not implemented or not supported/enabled in this service.
UNIMPLEMENTED = 12

"""
Internal errors.  Means some invariant expected by the underlying
system has been broken.  If you see one of these errors,
something is very broken.
"""
INTERNAL = 13

"""
The service is currently unavailable.  This is a most likely a
transient condition and may be corrected by retrying with
a backoff.

See litmus test above for deciding between FAILED_PRECONDITION,
ABORTED, and UNAVAILABLE.
"""
UNAVAILABLE = 14

# Unrecoverable data loss or corruption.
DATA_LOSS = 15

"""
The request does not have valid authentication credentials for the
operation.
"""
UNAUTHENTICATED = 16

"""
An extra enum entry to prevent people from writing code that
fails to compile when a new code is added.

Nobody should ever reference this enumeration entry. In particular,
if you write C++ code that switches on this enumeration, add a default:
case instead of a case that mentions this enumeration entry.

Nobody should rely on the value (currently 20) listed here.  It
may change in the future.
"""
DO_NOT_USE_RESERVED_FOR_FUTURE_EXPANSION_USE_DEFAULT_IN_SWITCH_INSTEAD_ = 20



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Validates responses and their security features."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from werkzeug import wrappers
from werkzeug.datastructures import Headers
from werkzeug import http

from tensorboard.util import tb_logging

# Loosely follow vocabulary from https://www.w3.org/TR/CSP/#framework-directives.
Directive = collections.namedtuple("Directive", ["name", "value"])

logger = tb_logging.get_logger()

_HTML_MIME_TYPE = "text/html"
_CSP_DEFAULT_SRC = "default-src"
# Whitelist of allowed CSP violations.
_CSP_IGNORE = {
    # Allow TensorBoard to be iframed.
    "frame-ancestors": ["*"],
    # Polymer-based code uses unsafe-inline.
    "style-src": ["'unsafe-inline'", "data:"],
    # Used in canvas
    "img-src": ["blob:", "data:"],
    # Used by numericjs.
    # TODO(stephanwlee): remove it eventually.
    "script-src": ["'unsafe-eval'"],
}


def _maybe_raise_value_error(error_msg):
    logger.warn("In 3.0, this warning will become an error:\n%s" % error_msg)
    # TODO(3.x): raise a value error.


class SecurityValidatorMiddleware(object):
    """WSGI middleware validating security on response.

    It validates:
    - responses have Content-Type
    - responses have X-Content-Type-Options: nosniff
    - text/html responses have CSP header. It also validates whether the CSP
      headers pass basic requirement. e.g., default-src should be present, cannot
      use "*" directive, and others. For more complete list, please refer to
      _validate_csp_policies.

    Instances of this class are WSGI applications (see PEP 3333).
    """

    def __init__(self, application):
        """Initializes an `SecurityValidatorMiddleware`.

        Args:
          application: The WSGI application to wrap (see PEP 3333).
        """
        self._application = application

    def __call__(self, environ, start_response):
        def start_response_proxy(status, headers, exc_info=None):
            self._validate_headers(headers)
            return start_response(status, headers, exc_info)

        return self._application(environ, start_response_proxy)

    def _validate_headers(self, headers_list):
        headers = Headers(headers_list)
        self._validate_content_type(headers)
        self._validate_x_content_type_options(headers)
        self._validate_csp_headers(headers)

    def _validate_content_type(self, headers):
        if headers.get("Content-Type"):
            return

        _maybe_raise_value_error("Content-Type is required on a Response")

    def _validate_x_content_type_options(self, headers):
        option = headers.get("X-Content-Type-Options")
        if option == "nosniff":
            return

        _maybe_raise_value_error(
            'X-Content-Type-Options is required to be "nosniff"'
        )

    def _validate_csp_headers(self, headers):
        mime_type, _ = http.parse_options_header(headers.get("Content-Type"))
        if mime_type != _HTML_MIME_TYPE:
            return

        csp_texts = headers.get_all("Content-Security-Policy")
        policies = []

        for csp_text in csp_texts:
            policies += self._parse_serialized_csp(csp_text)

        self._validate_csp_policies(policies)

    def _validate_csp_policies(self, policies):
        has_default_src = False
        violations = []

        for directive in policies:
            name = directive.name
            for value in directive.value:
                has_default_src = has_default_src or name == _CSP_DEFAULT_SRC

                if value in _CSP_IGNORE.get(name, []):
                    # There are cases where certain directives are legitimate.
                    continue

                # TensorBoard follows principle of least privilege. However, to make it
                # easier to conform to the security policy for plugin authors,
                # TensorBoard trusts request and resources originating its server. Also,
                # it can selectively trust domains as long as they use https protocol.
                # Lastly, it can allow 'none' directive.
                # TODO(stephanwlee): allow configuration for whitelist of domains for
                # stricter enforcement.
                # TODO(stephanwlee): deprecate the sha-based whitelisting.
                if (
                    value == "'self'"
                    or value == "'none'"
                    or value.startswith("https:")
                    or value.startswith("'sha256-")
                ):
                    continue

                msg = "Illegal Content-Security-Policy for {name}: {value}".format(
                    name=name, value=value
                )
                violations.append(msg)

        if not has_default_src:
            violations.append(
                "Requires default-src for Content-Security-Policy"
            )

        if violations:
            _maybe_raise_value_error("\n".join(violations))

    def _parse_serialized_csp(self, csp_text):
        # See https://www.w3.org/TR/CSP/#parse-serialized-policy.
        # Below Steps are based on the algorithm stated in above spec.
        # Deviations:
        # - it does not warn and ignore duplicative directive (Step 2.5)

        # Step 2
        csp_srcs = csp_text.split(";")

        policy = []
        for token in csp_srcs:
            # Step 2.1
            token = token.strip()

            if not token:
                # Step 2.2
                continue

            # Step 2.3
            token_frag = token.split(None, 1)
            name = token_frag[0]

            values = token_frag[1] if len(token_frag) == 2 else ""

            # Step 2.4
            name = name.lower()

            # Step 2.6
            value = values.split()
            # Step 2.7
            directive = Directive(name=name, value=value)
            # Step 2.8
            policy.append(directive)

        return policy
